# coding=utf-8
"""
Folders / Documents module
"""
from __future__ import absolute_import

from abilian.sbe.extension import sbe


def register_plugin(app):
    sbe.init_app(app)

    from .views import blueprint
    from .models import setup_listener
    from .commands import manager  # pylint: disable=bad-python3-import
    from . import signals  # noqa
    from . import lock

    app.register_blueprint(blueprint)
    setup_listener()

    # set default lock lifetime
    app.config.setdefault('SBE_LOCK_LIFETIME', lock.DEFAULT_LIFETIME)

    if app.script_manager:
        app.script_manager.add_command('documents', manager)



from __future__ import absolute_import, division, print_function, \
    unicode_literals

import pytest
from pytest import raises

from abilian.sbe.apps.forum.tests.util import get_email_message_from_file

from ..models import Post, Thread, ThreadClosedError
from ..tasks import process


def test_create_post():
    thread = Thread(title="Test thread")
    post = thread.create_post()
    assert post in thread.posts
    assert post.name == 'Test thread'

    thread.title = 'new title'
    assert thread.name == 'new title'
    assert post.name == 'new title'


def test_closed_property():
    thread = Thread(title='Test Thread')
    assert thread.closed is False
    thread.closed = True
    assert thread.closed is True
    thread.closed = 0
    assert thread.closed is False
    thread.closed = 1
    assert thread.closed is True
    assert thread.meta['abilian.sbe.forum']['closed'] is True


def test_thread_closed_guard():
    thread = Thread(title='Test Thread')
    thread.create_post()
    thread.closed = True

    with pytest.raises(ThreadClosedError):
        thread.create_post()

    p = Post(body_html='ok')

    with pytest.raises(ThreadClosedError):
        p.thread = thread

    thread.closed = False
    p.thread = thread
    assert len(thread.posts) == 2

    thread.closed = True
    with pytest.raises(ThreadClosedError):
        del thread.posts[0]

    assert len(thread.posts) == 2

    with pytest.raises(ThreadClosedError):
        # actually thread.posts will be replaced by `[]` and we can't prevent
        # this, but exception has been raised
        thread.posts = []

    thread.closed = False
    thread.posts = [p]
    thread.closed = True
    assert thread.posts == [p]
    assert p.thread is thread
    with pytest.raises(ThreadClosedError):
        p.thread = None


def test_change_thread_copy_name():
    thread = Thread(title='thread 1')
    thread2 = Thread(title='thread 2')
    post = Post(thread=thread, body_html='post content')
    assert post.name == thread.name

    post.thread = thread2
    assert post.name == thread2.name


def test_task_process_email():
    """Test the process_email function."""

    marker = '_____Write above this line to post_____'

    message = get_email_message_from_file('reply.email')
    newpost = process(message, marker)[0]
    assert newpost

    message = get_email_message_from_file('reply_nocharset_specified.email')
    newpost = process(message, marker)[0]
    assert newpost

    message = get_email_message_from_file('reply_no_marker.email')
    with raises(LookupError):
        process(message, marker)

    # dubious check
    message = get_email_message_from_file('reply_no_textpart.email')
    with raises(LookupError):
        process(message, marker)



"""
Parses XML messages and converts them to objects.

"""
from __future__ import absolute_import, print_function, unicode_literals

import base64
from datetime import datetime

from lxml import objectify

ATOM_NS = "http://www.w3.org/2005/Atom"
APP_NS = "http://www.w3.org/2007/app"
CMISRA_NS = "http://docs.oasis-open.org/ns/cmis/restatom/200908/"
CMIS_NS = "http://docs.oasis-open.org/ns/cmis/core/200908/"


class Entry(object):

    def __init__(self, xml=None):
        self.properties = {}
        self.links = []
        self.content = None
        self.content_type = None

        if xml:
            self.parse(xml)

    def parse(self, xml):
        root = objectify.fromstring(xml)
        object = root['{%s}object' % CMISRA_NS]
        properties = object['{%s}properties' % CMIS_NS]
        for element in properties.iterchildren():
            property = Property(element)
            self.properties[property.property_definition_id] = property

        content_element = getattr(root, '{%s}content' % CMISRA_NS, None)
        if content_element is not None:
            self.content_type = content_element.mediatype.text
            self.content = base64.b64decode(content_element.base64.text)

    @property
    def name(self):
        return self.properties['cmis:name'].value

    @property
    def type(self):
        return self.properties['cmis:objectTypeId'].value


class Property(object):
    """
    A property MAY hold zero, one, or more typed data value(s).
    Each property MAY be single-valued or multi-valued. A single-valued property
    contains a single data value, whereas a multi-valued property contains an
    ordered list of data values of the same type. The ordering of values in a
    multi-valued property SHOULD be preserved by the repository.

    A property, either single-valued or multi-valued, MAY be in a "not set" state.
    CMIS does not support "null" property value. If a multi-valued property
    is not in a "not set" state, its property value MUST be a non-empty list of
    individual values. Each individual value in the list MUST NOT be in a "not
    set" state and MUST conform to the property's property-type.

    A multi-valued property is either set or not set in its entirety. An
    individual value of a multi-valued property MUST NOT be in an individual
    "value not set" state and hold a position in the list of values. An empty
    list of values MUST NOT be allowed.

    Every property is typed. The property-type defines the data type of the data
    value(s) held by the property. CMIS specifies the following property-types.
    """

    def __init__(self, element=None):
        if element is not None:
            self.parse(element)

    def parse(self, element):
        tag = element.tag
        self.type = tag[tag.index("}") + 1 + len("property"):].lower()
        self.property_definition_id = element.attrib['propertyDefinitionId']
        self.local_name = element.attrib.get('localName')
        self.display_name = element.attrib.get('displayName')
        self.query_name = element.attrib.get('queryName')

        value_elem = getattr(element, "{%s}value" % CMIS_NS)
        if value_elem:
            value = value_elem.text
        else:
            self.value = value = None

        if value:
            if self.type in ('id', 'string'):
                self.value = value
            elif self.type == 'datetime':
                # FIXME
                self.value = datetime(value)
            else:
                raise Exception("Unknown value type: %s" % self.type)



from __future__ import absolute_import, print_function, unicode_literals

from abilian.core.models.subjects import User
from abilian.sbe.apps.communities.models import READER, Community
from abilian.sbe.testing import BaseTestCase


class CommunityBaseTestCase(BaseTestCase):

    no_login = True

    def setUp(self):
        super(CommunityBaseTestCase, self).setUp()
        self.community = Community(name="My Community")
        self.session.add(self.community)
        self.session.flush()


class CommunityIndexingTestCase(CommunityBaseTestCase):
    """
    Testcase for testing indexing and searching of community content.
    """

    no_login = False
    SERVICES = ('security', 'indexing')

    def setUp(self):
        super(CommunityIndexingTestCase, self).setUp()

        self.svc = self.app.services['indexing']

        self.user = User(
            email='user_1@example.com', password='azerty', can_login=True)
        self.session.add(self.user)
        self.community.set_membership(self.user, READER)

        self.c2 = Community(name='Other community')
        self.session.add(self.c2)

        self.user_c2 = User(
            email='user_2@example.com', password='azerty', can_login=True)
        self.session.add(self.user_c2)
        self.c2.set_membership(self.user_c2, READER)

        self.user_no_community = User(
            email='no_community@example.com', password='azerty', can_login=True)
        self.session.add(self.user_no_community)
        self.session.commit()



#!/usr/bin/env python
# coding=utf-8
"""
"""

from __future__ import absolute_import, print_function, unicode_literals

import logging

from flask_script import Manager

from abilian.core.commands import setup_abilian_commands
from abilian.sbe.app import Application


def create_app():
    return Application(name='myapp')


if __name__ == '__main__':
    logging.basicConfig()
    logging.getLogger().setLevel(logging.INFO)
    manager = Manager(create_app)
    setup_abilian_commands(manager)

    manager.run()



# encoding: utf-8
"""
Test the application using py.test and splinter.

See: https://splinter.readthedocs.org/

Also, we're using pytest-splinter
(https://github.com/pytest-dev/pytest-splinter) to inject the browser as
a pytest fixture.
"""
from __future__ import absolute_import, print_function, unicode_literals

import multiprocessing
import shutil
import socket
import tempfile

import pytest
from werkzeug.serving import select_ip_version

from abilian.sbe.app import Application


@pytest.fixture(scope='session')
def app_port(request):
    port = 30000
    hostname = '127.0.0.1'

    address_family = select_ip_version(hostname, port)
    test_socket = socket.socket(address_family, socket.SOCK_STREAM)
    test_socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    while port < 60000:
        try:
            test_socket.bind((hostname, port))
            test_socket.close()

        except socket.error as exc:
            if exc.errno != 98:
                # errno(98, Address already in use)
                raise
            port += 1

        break

    return port


@pytest.fixture(scope='session')
def app_root(app_port):
    return "http://localhost:{}".format(app_port)


# We're using phantomjs as our default browser.
@pytest.fixture(scope='session')  # pragma: no cover
def splinter_webdriver(request):
    return 'phantomjs'


@pytest.fixture(scope='session')
def splinter_driver_kwargs():
    """Webdriver kwargs."""
    # set resourceTimeout in the hope it will help jenkins kill phantomjs after
    # tests
    return dict(
        desired_capabilities={
            'phantomjs.page.settings.resourceTimeout': '30000',  # 30s
        },)


@pytest.fixture(scope='module')
def instance_path(request):
    """Creates a temporary directory for instance data.
    """
    tmp_dir = tempfile.mkdtemp(prefix='tmp-pytest-', suffix='-abilian-sbe')

    def clear():
        shutil.rmtree(tmp_dir)

    request.addfinalizer(clear)
    return tmp_dir


@pytest.fixture(scope='module')
def app(request, instance_path, app_port):
    app = Application(instance_path=instance_path)

    # FIXME: need a working environment (DB, Redis...)
    # with app.app_context():
    #   app.create_db()

    def worker(app, port):
        app.run(port=port)

    process = multiprocessing.Process(target=worker, args=(app, app_port))

    try:
        process.start()
    except Exception as e:
        pytest.fail(e.message)

    def finalizer():
        if process:
            process.terminate()

    request.addfinalizer(finalizer)

    return app


"""
@skip
def test_home(browser, app, app_root):
    browser.visit(app_root)


@skip
def test_login(browser, app, app_root):
    browser.visit(app_root + '/user/login')
    browser.fill('email', "admin@example.com")
    browser.fill('password', "admin")
    button = browser.find_by_xpath("//form[@name='login']//button")
    assert button

    # FIXME: at this point this raises an error because we have to set up the db.
    # button.click()
    # assert "Welcome to Abilian" in browser.find_element_by_xpath("/html/body").text


@skip
def test_forgotten_pw(browser, app, app_root):
    browser.visit(app_root + '/user/forgotten_pw')
"""



from __future__ import absolute_import, print_function, unicode_literals

from flask_babel import lazy_gettext as _l

from abilian.core.util import BasePresenter


class CommunityPresenter(BasePresenter):

    @property
    def breadcrumbs(self):
        return [{
            'label': _l("Communities"),
            'path': "/communities/"
        }, {
            'label': self._model.name
        }]



# coding=utf-8
from __future__ import absolute_import, print_function, unicode_literals

import re

from flask import g, url_for
from markdown import Markdown
from mock import MagicMock, patch
from six import text_type
from six.moves.urllib.parse import quote_plus
from toolz import first

from abilian.core.models.subjects import User
from abilian.sbe.apps.communities.tests.base import CommunityBaseTestCase, \
    CommunityIndexingTestCase
from abilian.sbe.apps.wiki.markup import SBEWikiLinkExtension

from . import views
from .models import WikiPage


def test_wikilink_extension():
    text = '/*€('
    wikilink = '[[' + text + ']]'

    def build_url(label, base, end):
        return '/?title=' + quote_plus(label.encode('utf-8')) + end

    extension = SBEWikiLinkExtension([('build_url', build_url)])
    ctx = {}
    ctx['extensions'] = [extension, 'toc']
    ctx['output_format'] = 'html5'
    md = Markdown(**ctx)

    page_exists_mock = MagicMock(return_value=True)
    with patch('abilian.sbe.apps.wiki.forms.page_exists', page_exists_mock):
        result = md.convert(wikilink)

    qtext = text_type(quote_plus(text.encode('utf-8')))
    expected = '<p><a class="wikilink" href="/?title={href}/">{text}</a></p>'
    expected = expected.format(href=qtext, text=text)
    assert expected == result

    # make sur the class is 'wikilink new' if page exists'
    page_exists_mock = MagicMock(return_value=False)
    with patch('abilian.sbe.apps.wiki.forms.page_exists', page_exists_mock):
        result = md.convert(wikilink)

    qtext = text_type(quote_plus(text.encode('utf-8')))
    expected = '<p><a class="wikilink new" href="/?title={href}/">{text}</a></p>'
    expected = expected.format(href=qtext, text=text)
    assert expected == result


class WikiBaseTestCase(CommunityBaseTestCase):

    def setUp(self):
        super(WikiBaseTestCase, self).setUp()
        g.user = User.query.all()[0]


class TestModels(WikiBaseTestCase):

    def test_create_page(self):
        page = WikiPage(title="Some page", body_src='abc')
        assert page.title == 'Some page'
        assert page.name == 'Some page'
        assert page.body_src == 'abc'
        assert page.body_html == '<p>abc</p>'
        assert len(page.revisions) == 1

        revision = page.revisions[0]
        assert revision.number == 0
        assert revision.author == g.user

    def test_rename_page(self):
        page = WikiPage(title="Some page", body_src='abc')
        assert page.title == 'Some page'
        assert page.name == 'Some page'

        page.title = 'Title Renamed'
        assert page.title == 'Title Renamed'
        assert page.name == 'Title Renamed'

        page.name = 'Name'
        assert page.title == 'Name'
        assert page.name == 'Name'

    def test_create_revision(self):
        page = WikiPage('abc')
        page.create_revision("def", "page updated")

        assert len(page.revisions) == 2
        last_revision = page.revisions[1]
        assert last_revision.number == 1
        assert last_revision.author == g.user


class TestIndexing(CommunityIndexingTestCase):

    def setUp(self):
        super(TestIndexing, self).setUp()
        g.user = User.query.all()[0]

    def test_wiki_indexed(self):
        page = WikiPage(title='Community 1', community=self.community)
        self.session.add(page)
        page_other = WikiPage(title='Community 2: other', community=self.c2)
        self.session.add(page_other)
        self.session.commit()

        svc = self.svc
        obj_types = (WikiPage.entity_type,)
        with self.login(self.user_no_community):
            res = svc.search('community', object_types=obj_types)
            assert len(res) == 0

        with self.login(self.user):
            res = svc.search('community', object_types=obj_types)
            assert len(res) == 1

            hit = res[0]
            assert hit['object_key'] == page.object_key

        with self.login(self.user_c2):
            res = svc.search('community', object_types=obj_types)
            assert len(res) == 1

            hit = res[0]
            assert hit['object_key'] == page_other.object_key


class TestsViews(WikiBaseTestCase):

    def test_home(self):
        response = self.client.get(
            url_for("wiki.index", community_id=self.community.slug))
        assert response.status_code == 302

        response = self.client.get(
            url_for(
                "wiki.page", title='Home', community_id=self.community.slug))
        assert response.status_code == 200

    def test_create_page_initial_form(self):
        g.community = self.community
        view = views.PageCreate()
        view.prepare_args([], {})
        form = view.form
        assert form['last_revision_id'].data is None

    def test_create_page(self):
        title = 'Some page name'
        url = url_for("wiki.page_new", community_id=self.community.slug)
        url += '?title=Some+page+name'
        response = self.client.get(url)
        assert response.status_code == 200
        # make sure the title is filled when comming from wikilink
        assert 'value="Some page name"' in response.get_data(as_text=True)

        title = 'Some page'
        body = "LuuP3jai"
        url = url_for("wiki.page_new", community_id=self.community.slug)
        data = dict(title=title, body_src=body, __action="create")
        response = self.client.post(url, data=data)
        assert response.status_code == 302

        url = url_for(
            "wiki.page", community_id=self.community.slug, title=title)
        response = self.client.get(url)
        assert response.status_code == 200
        assert body in response.get_data(as_text=True)

        # edit
        url = url_for(
            "wiki.page_edit", community_id=self.community.slug, title=title)
        response = self.client.get(url)
        assert response.status_code == 200

        # Slightly hackish way to get the page_id
        line = first(
            l for l in response.get_data(as_text=True).split("\n")
            if 'name="page_id"' in l)
        m = re.search('value="(.*?)"', line)
        page_id = int(m.group(1))

        url = url_for(
            "wiki.page_changes", community_id=self.community.slug, title=title)
        response = self.client.get(url)
        assert response.status_code == 200

        url = url_for(
            "wiki.page_source", community_id=self.community.slug, title=title)
        response = self.client.get(url)
        assert response.status_code == 200

        url = url_for("wiki.page_edit", community_id=self.community.slug)
        data = dict(title=title, page_id=page_id, body_src="abc def")
        data['__action'] = 'edit'
        response = self.client.post(url, data=data)
        assert response.status_code == 302

        url = url_for(
            "wiki.page_compare",
            rev0="on",
            rev1="on",
            community_id=self.community.slug,
            title=title)
        response = self.client.get(url)
        assert response.status_code == 200



# -*- coding: utf-8 -*-
"""
Document management blueprint.
"""
from __future__ import absolute_import, print_function, unicode_literals

from flask import g

from abilian.i18n import _l
from abilian.sbe.apps.communities.blueprint import Blueprint
from abilian.sbe.apps.communities.security import is_manager
from abilian.web.action import Endpoint
from abilian.web.nav import BreadcrumbItem

from ..actions import register_actions

__all__ = ['blueprint']

blueprint = Blueprint(
    "documents", __name__, url_prefix="/docs", template_folder="../templates")
route = blueprint.route
blueprint.record_once(register_actions)


@blueprint.url_value_preprocessor
def init_document_values(endpoint, values):
    g.current_tab = 'documents'
    g.is_manager = is_manager()

    g.breadcrumb.append(
        BreadcrumbItem(
            label=_l('Documents'),
            url=Endpoint('documents.index', community_id=g.community.slug)))



# coding=utf-8
"""
Lightweight integration and denormalisation using events (signals).
"""

from __future__ import absolute_import, print_function, unicode_literals

from blinker import ANY

from abilian.core.signals import activity
from abilian.sbe.apps.documents.models import Document

from .models import Community


@activity.connect_via(ANY)
def update_community(sender, verb, actor, object, target=None):
    if isinstance(object, Community):
        object.touch()
        return

    if isinstance(target, Community):
        community = target
        community.touch()

        if isinstance(object, Document):
            if verb == 'post':
                community.document_count += 1
            elif verb == 'delete':
                community.document_count -= 1



# coding=utf-8

from __future__ import absolute_import, print_function, unicode_literals

from flask_babel import lazy_gettext as _l
from wtforms import StringField, TextAreaField
from wtforms_alchemy import model_form_factory

from abilian.sbe.apps.communities.models import Community
from abilian.web.forms import Form
from abilian.web.forms import widgets as abilian_widgets
from abilian.web.forms.fields import QuerySelect2Field
from abilian.web.forms.filters import strip
from abilian.web.forms.validators import optional, required

ModelForm = model_form_factory(Form)


# FIXME
class UserProfileForm(ModelForm):
    pass


class UserProfileViewForm(UserProfileForm):
    communautes = QuerySelect2Field(
        'Communautés d\'appartenance',
        get_label='name',
        view_widget=abilian_widgets.ListWidget(),
        query_factory=lambda: Community.query.all(),
        multiple=True,
        validators=[optional()])


class GroupForm(Form):
    name = StringField(
        _l("Name"),
        filters=(strip,),
        validators=[required(message=_l("Name is required."))])

    description = TextAreaField(_l("Description"))



# coding=utf-8
"""
Tests from test_community are currently refactored using pytest in this module.
"""
from __future__ import absolute_import, division, print_function, \
    unicode_literals

import pytest
import six
import sqlalchemy as sa
from flask_login import login_user, logout_user
from mock import mock
from pytest import fixture

from abilian.core.entities import Entity
from abilian.core.models.subjects import User
from abilian.sbe.apps.documents.models import Folder

from .. import signals, views
from ..models import MEMBER, READER, Community, CommunityIdColumn, \
    community_content


@fixture
def community(db):
    community = Community(name="My Community")
    db.session.add(community)
    db.session.commit()
    return community


#
# Actual tests
#
def test_instanciation(db):
    community = Community(name="My Community")
    assert isinstance(community.folder, Folder)
    # assert isinstance(community.group, Group)


def test_default_view_kw():
    # test exceptions are handled if passed an object with 'community' attribute
    # and no community_id in kwargs. and ValueError is properly raised
    if six.PY2:
        dummy = type(b'Dummy', (object,), {b'community': None})()
    else:
        dummy = type('Dummy', (object,), {'community': None})()

    with pytest.raises(ValueError) as exc_info:
        views.default_view_kw({}, dummy, 'dummy', 1)

    if six.PY2:
        assert exc_info.value.message == 'Cannot find community_id value'
    else:
        assert exc_info.value.args == ('Cannot find community_id value',)


def test_default_url(app, community):
    url = app.default_view.url_for(community)
    assert url.endswith('/localhost/communities/my-community/')


def test_can_recreate_with_same_name(community, db):
    name = community.name

    db.session.delete(community)
    db.session.commit()

    community = Community(name=name)
    db.session.add(community)

    # if community.folder was not deleted, this will raise IntegrityError. Test
    # passes if no exceptions is raised
    db.session.commit()


def test_rename(community):
    NEW_NAME = "My new name"
    community.rename(NEW_NAME)
    assert community.name == NEW_NAME
    assert community.folder.name == NEW_NAME


def test_auto_slug(community):
    assert community.slug == 'my-community'


def test_membership(community, db):
    user = User(email="user@example.com")

    memberships = community.memberships
    assert memberships == []

    # setup signals testers with mocks.
    when_set = mock.MagicMock()
    when_set.mock_add_spec(['__name__'])  # required for signals
    signals.membership_set.connect(when_set)
    when_removed = mock.MagicMock()
    when_removed.mock_add_spec(['__name__'])
    signals.membership_removed.connect(when_removed)

    # invalid role
    with pytest.raises(ValueError):
        community.set_membership(user, 'dummy role name')

    assert not when_set.called
    assert not when_removed.called

    # simple member
    community.set_membership(user, "member")
    db.session.commit()

    memberships = community.memberships
    assert len(memberships) == 1
    assert memberships[0].user == user
    assert memberships[0].role is MEMBER

    when_set.assert_called_once_with(
        community, is_new=True, membership=memberships[0])
    assert not when_removed.called
    when_set.reset_mock()

    assert community.get_role(user) is MEMBER

    assert community.get_memberships() == [memberships[0]]
    assert community.get_memberships('member') == [memberships[0]]
    assert community.get_memberships('manager') == []

    # change user role
    community.set_membership(user, "manager")
    db.session.commit()

    memberships = community.memberships
    assert len(memberships) == 1
    assert memberships[0].user == user
    assert memberships[0].role == "manager"

    assert community.get_role(user) == "manager"

    when_set.assert_called_once_with(
        community, is_new=False, membership=memberships[0])
    assert not when_removed.called
    when_set.reset_mock()

    # remove user
    membership = memberships[0]
    community.remove_membership(user)
    db.session.commit()

    memberships = community.memberships
    assert memberships == []

    assert not when_set.called
    when_removed.assert_called_once_with(community, membership=membership)


def test_folder_roles(community, db, app):
    user = User(email="user@example.com")
    folder = community.folder
    community.set_membership(user, "member")
    db.session.commit()
    security = app.services['security']

    assert security.get_roles(user, folder) == ["reader"]

    # this tests a bug, where local roles whould disappear when setting
    # membership twice
    community.set_membership(user, "member")
    assert security.get_roles(user, folder) == ["reader"]


def test_community_content_decorator(community, db):

    @community_content
    class CommunityContent(Entity):
        community_id = CommunityIdColumn()
        community = sa.orm.relation(Community, foreign_keys=[community_id])

    sa.orm.configure_mappers()
    conn = db.session.connection()
    for table in sa.inspect(CommunityContent).tables:
        if not table.exists(conn):
            table.create(conn)

    cc = CommunityContent(name='my content', community=community)
    db.session.add(cc)
    db.session.flush()
    assert hasattr(cc, 'community_slug')
    assert cc.community_slug == 'my-community'
    assert cc.slug == 'my-content'

    index_to = dict(CommunityContent.__indexation_args__['index_to'])
    assert 'community_slug' in index_to


################################################################################


def login(user, remember=False, force=False):
    """Perform user login for `user`, so that code needing a logged-in user
    can work.

    This method can also be used as a context manager, so that logout is
    performed automatically::

        with login(user):
            assert ...

    .. seealso:: :meth:`logout`
    """
    # self._login_tests_sanity_check()
    success = login_user(user, remember=remember, force=force)
    if not success:
        raise ValueError('User is not active, cannot login; or use force=True',)

    class LoginContext(object):

        def __enter__(self):
            return None

        def __exit__(self, type, value, traceback):
            logout()

    return LoginContext()


def logout():
    logout_user()


def test_community_indexed(app, db):
    index_service = app.services['indexing']
    index_service.start()

    security_service = app.services['security']
    security_service.start()

    obj_types = (Community.entity_type,)

    user_no_community = User(email='no_community@example.com')
    db.session.add(user_no_community)

    community1 = Community(name="My Community")
    db.session.add(community1)

    community2 = Community(name='Other community')
    db.session.add(community2)

    user = User(email='user_1@example.com')
    db.session.add(user)
    community1.set_membership(user, READER)

    user_c2 = User(email='user_2@example.com')
    db.session.add(user_c2)
    community2.set_membership(user_c2, READER)

    db.session.commit()

    with app.test_request_context():
        with login(user_no_community):
            res = index_service.search('community', object_types=obj_types)
            assert len(res) == 0

        with login(user):
            res = index_service.search('community', object_types=obj_types)
            assert len(res) == 1
            hit = res[0]
            assert hit['object_key'] == community1.object_key

        with login(user_c2):
            res = index_service.search('community', object_types=obj_types)
            assert len(res) == 1
            hit = res[0]
            assert hit['object_key'] == community2.object_key


def test_default_view_kw_with_hit(app, db, community):
    index_service = app.services['indexing']
    index_service.start()

    security_service = app.services['security']
    security_service.start()

    user = User(email='user_1@example.com')
    db.session.add(user)
    community.set_membership(user, READER)

    obj_types = (Community.entity_type,)

    with app.test_request_context():
        with login(user):
            hit = index_service.search('community', object_types=obj_types)[0]
            kw = views.default_view_kw({}, hit, hit['object_type'], hit['id'])
            assert kw == {'community_id': community.slug}



from __future__ import absolute_import


def register_plugin(app):
    from .views import wiki
    from .actions import register_actions

    wiki.record_once(register_actions)
    app.register_blueprint(wiki)



from __future__ import absolute_import, print_function, unicode_literals

from unittest import skip

from flask import url_for

from abilian.sbe.testing import BaseTestCase


class TestsViews(BaseTestCase):
    no_login = True

    @skip("Doesn't work. Needs a real user.")
    def test_crm_notifications(self):
        response = self.client.get(url_for("preferences.crm_notifications"))
        assert response.status_code == 200



# coding=utf-8
"""
This panel manages user setting for email reminders related to the SBE
(social netowking) app.
"""
from __future__ import absolute_import, print_function, unicode_literals

from flask import current_app as app
from flask import flash, redirect, render_template, request, url_for
from werkzeug.exceptions import InternalServerError
from wtforms import BooleanField

from abilian.core.extensions import db
from abilian.i18n import _, _l
from abilian.services.preferences.panel import PreferencePanel
from abilian.web import csrf
from abilian.web.forms import Form, widgets


class SbeNotificationsForm(Form):
    daily = BooleanField(
        label=_("Receive by email a daily digest of activities in your communities"),
        widget=widgets.BooleanWidget(on_off_mode=True))


class SbeNotificationsPanel(PreferencePanel):
    id = 'sbe_notifications'
    label = _l('Community notifications')

    def is_accessible(self):
        return True

    def get(self):
        # Manual security check, should be done by the framework instead.
        if not self.is_accessible():
            raise InternalServerError()

        preferences = app.services['preferences']
        data = {}
        prefs = preferences.get_preferences()

        for k, v in prefs.items():
            if k.startswith("sbe:notifications:"):
                data[k[18:]] = v

        form = SbeNotificationsForm(formdata=None, prefix=self.id, **data)
        return render_template("preferences/sbe_notifications.html", form=form)

    @csrf.protect
    def post(self):
        # Manual security check, should be done by the framework instead.
        if not self.is_accessible():
            raise InternalServerError()

        if request.form['_action'] == 'cancel':
            return redirect(url_for(".sbe_notifications"))
        form = SbeNotificationsForm(request.form, prefix=self.id)
        if form.validate():
            preferences = app.services['preferences']
            for field_name, field in form._fields.items():
                if field is form.csrf_token:
                    continue
                key = 'sbe:notifications:{}'.format(field_name)
                value = field.data
                preferences.set_preferences(**{key: value})

            db.session.commit()
            flash(_("Preferences saved."), "info")
            return redirect(url_for(".sbe_notifications"))
        else:
            return render_template(
                "preferences/sbe_notifications.html", form=form)



# coding=utf-8
"""
"""
from __future__ import absolute_import, print_function, unicode_literals

from flask import Markup, render_template


class UserPhotoInputWidget(object):

    def __call__(self, field, **kwargs):
        kwargs.setdefault('id', field.id)
        user_id = None

        if 'widget_options' in kwargs:
            options = kwargs.pop('widget_options')
            user_id = options.get('user_id')

        return Markup(
            render_template(
                'social/widget_user_photo.html',
                field=field,
                attrs=kwargs,
                user_id=user_id))



# coding=utf-8
"""
"""
from __future__ import absolute_import, print_function, unicode_literals

import fnmatch
import itertools
import logging
import os
import re
import tempfile
from datetime import datetime
from functools import partial
from io import StringIO
from typing import Any, List
from zipfile import ZipFile, is_zipfile

import sqlalchemy as sa
import whoosh.query as wq
from flask import Markup, current_app, flash, g, jsonify, make_response, \
    redirect, render_template, render_template_string, request, send_file, \
    session
from six import text_type
from six.moves.urllib.parse import quote
from sqlalchemy import func
from werkzeug.exceptions import InternalServerError
from xlwt import Workbook, easyxf

from abilian.core.extensions import db
from abilian.core.models.subjects import Group, User
from abilian.core.signals import activity
from abilian.i18n import _, _n
from abilian.sbe.apps.communities.views import default_view_kw
from abilian.services import get_service
from abilian.services.security import READ, WRITE, Role, security
from abilian.web import csrf, http, url_for
from abilian.web.action import actions
from abilian.web.views import default_view

from ..models import Document, Folder, icon_for, icon_url
from ..repository import repository
from ..search import reindex_tree
from .util import breadcrumbs_for, check_manage_access, check_read_access, \
    check_write_access, create_document, edit_object, get_document, \
    get_folder, get_new_filename, get_selected_objects
from .views import blueprint

route = blueprint.route

__all__ = ()


@route("/")
def index():
    folder = g.community.folder
    url = url_for(folder)
    return redirect(url)


@default_view(blueprint, Folder, id_attr='folder_id', kw_func=default_view_kw)
@route("/folder/<int:folder_id>")
def folder_view(folder_id):
    folder = get_folder(folder_id)
    bc = breadcrumbs_for(folder)
    actions.context['object'] = folder
    ctx = {
        'folder': folder,
        'children': folder.filtered_children,
        'breadcrumbs': bc,
    }

    view_style = session.get('sbe_doc_view_style', 'thumbnail_view')
    if view_style == 'thumbnail_view':
        resp = render_template("documents/folder.html", **ctx)
    elif view_style == 'gallery_view':
        resp = render_template("documents/folder_gallery_view.html", **ctx)
    else:
        raise InternalServerError("Unknown value for sbe_doc_view_style")

    return resp


@route("/folder/change_view_style/<int:folder_id>", methods=["GET", "POST"])
@csrf.protect
def change_view_style(folder_id):
    folder = get_folder(folder_id)
    if request.method == "POST":
        view_style = request.form["view_style"]
        if view_style == "gallery_view":
            session['sbe_doc_view_style'] = 'gallery_view'
        else:
            session['sbe_doc_view_style'] = 'thumbnail_view'
        return redirect(
            url_for(
                ".folder_view",
                folder_id=folder_id,
                community_id=folder.community.slug))
    else:
        return redirect(
            url_for(
                ".folder_view",
                folder_id=folder_id,
                community_id=folder.community.slug))


@route("/folder/<int:folder_id>/json")
def folder_json(folder_id):
    """Return parent folder + subfolders."""

    folder = get_folder(folder_id)
    folder_url = partial(url_for, '.folder_json')
    result = {}
    has_permission = security.has_permission
    result['current_folder_selectable'] = has_permission(
        g.user, WRITE, folder, inherit=True)
    folders = result['folders'] = []
    bc = result['breadcrumbs'] = []
    subfolders = sorted(
        (f for f in folder.subfolders
         if has_permission(g.user, READ, f, inherit=True)),
        key=lambda f: f.title)

    parent = folder

    # breadcrumbs
    for parent in reversed(list(folder._iter_to_root())):
        if parent.is_root_folder:
            continue

        data = {
            'id':
            parent.id,
            'url':
            folder_url(folder_id=parent.id, community_id=parent.community.slug),
            'title':
            parent.title,
        }
        bc.append(data)

    if folder.parent is not None and not folder.parent.is_root_folder:
        # not at root folder: allow to go 1 level up
        data = bc[-2].copy()
        data['title'] = ".."
        folders.append(data)
    else:
        result['current_folder_selectable'] = False

    for folder in subfolders:
        data = {
            'id':
            folder.id,
            'url':
            folder_url(folder_id=folder.id, community_id=folder.community.slug),
            'title':
            folder.title,
        }
        folders.append(data)

    return jsonify(result)


@route("/folder/<int:folder_id>/members")
def members(folder_id):
    folder = get_folder(folder_id)
    bc = breadcrumbs_for(folder)
    actions.context['object'] = folder
    members = folder.members()
    ctx = {
        'folder': folder,
        'members': members,
        'breadcrumbs': bc,
    }
    return render_template("community/members.html", **ctx)


@route("/folder/<int:folder_id>/permissions")
@http.nocache
def permissions(folder_id):
    folder = get_folder(folder_id)
    check_manage_access(folder)

    bc = breadcrumbs_for(folder)
    actions.context['object'] = folder
    local_roles_assignments = folder.get_local_roles_assignments()
    principals = {p for p, r in local_roles_assignments}
    security._fill_role_cache_batch(principals)

    users_and_local_roles = [(user, role, repository.has_access(user, folder))
                             for user, role in local_roles_assignments
                             if isinstance(user, User)]
    groups_and_local_roles = [
        t for t in local_roles_assignments if isinstance(t[0], Group)
    ]

    users_and_inherited_roles = groups_and_inherited_roles = ()

    if folder.inherit_security:
        inherited_roles_assignments = folder.get_inherited_roles_assignments()
        users_and_inherited_roles = [
            (user, role, False)
            for user, role in inherited_roles_assignments
            if isinstance(user, User)
        ]
        groups_and_inherited_roles = [
            t for t in inherited_roles_assignments if isinstance(t[0], Group)
        ]

    query = Group.query
    query = query.order_by(func.lower(Group.name))
    all_groups = query.all()

    class EntryPresenter(object):
        _USER_FMT = ('<a href="{{ url_for("social.user", user_id=user.id) }}">'
                     '{{ user.name }}</a>')
        _GROUP_FMT = (
            '<a href="{{ url_for("social.group_home", group_id=group.id)'
            ' }}">{{ group.name }}</a>')

        def __init__(self, e):
            render = render_template_string
            self.entry = e
            self.date = e.happened_at.strftime('%Y-%m-%d %H:%M')
            self.manager = render(
                '<img src="{{ user_photo_url(e.manager, size=16) }}" alt="" />'
                '<a href="{{ url_for("social.user", user_id=e.manager_id) }}">'
                '{{ e.manager.name }}</a>',
                e=e)

            if e.op == e.SET_INHERIT:
                msg = _('On {date}, {manager} has activated inheritance')
            elif e.op == e.UNSET_INHERIT:
                msg = _('On {date}, {manager} has deactivated inheritance')
            elif e.op == e.GRANT:
                msg = _('On {date}, {manager} has given role "{role}" to {principal}')
            elif e.op == e.REVOKE:
                msg = _('On {date}, {manager} has revoked role "{role}" from '
                        '{principal}')
            else:
                raise Exception("Unknown audit entry type %s" % e.op)

            principal = ''
            if self.entry.user:
                principal = render(self._USER_FMT, user=self.entry.user)
            elif self.entry.group:
                principal = render(self._GROUP_FMT, group=self.entry.group)

            self.msg = Markup(
                msg.format(
                    date=self.date,
                    manager=self.manager,
                    role=self.entry.role,
                    principal=principal))

    audit_entries = [EntryPresenter(e) for e in security.entries_for(folder)]

    ctx = {
        'folder': folder,
        'users_and_local_roles': users_and_local_roles,
        'users_and_inherited_roles': users_and_inherited_roles,
        'groups_and_local_roles': groups_and_local_roles,
        'groups_and_inherited_roles': groups_and_inherited_roles,
        'audit_entries': audit_entries,
        'all_groups': all_groups,
        'breadcrumbs': bc,
    }
    return render_template("documents/permissions.html", **ctx)


@route("/folder/<int:folder_id>/permissions", methods=['POST'])
@csrf.protect
def permissions_update(folder_id):
    folder = repository.get_folder_by_id(folder_id)
    check_manage_access(folder)
    has_permission = security.has_permission
    action = request.form.get("action")

    if action in ("activate_inheritance", "deactivate_inheritance"):
        inherit_security = (action == "activate_inheritance")

        if not (inherit_security or
                has_permission(g.user, 'manage', folder, inherit=False)):
            # don't let user shoot himself in the foot
            flash(_('You must have the "manager" local role on this folder in '
                    'order to deactivate inheritance.'), 'error')
            return redirect(
                url_for(
                    ".permissions",
                    folder_id=folder_id,
                    community_id=folder.community.slug))

        security.set_inherit_security(folder, inherit_security)
        db.session.add(folder)
        reindex_tree(folder)
        db.session.commit()
        return redirect(
            url_for(
                ".permissions",
                folder_id=folder_id,
                community_id=folder.community.slug))

    elif action == "add-user-role":
        role = request.form.get("role").lower()
        user_id = int(request.form.get("user"))
        user = User.query.get(user_id)

        security.grant_role(user, role, folder)
        reindex_tree(folder)
        db.session.commit()
        return redirect(
            url_for(
                ".permissions",
                folder_id=folder_id,
                community_id=folder.community.slug))

    elif action == "add-group-role":
        role = request.form.get("role").lower()
        group_id = int(request.form.get("group"))
        group = Group.query.get(group_id)

        security.grant_role(group, role, folder)
        reindex_tree(folder)
        db.session.commit()
        return redirect(
            url_for(
                ".permissions",
                folder_id=folder_id,
                community_id=folder.community.slug))

    else:
        action, args = request.form.items()[0]
        role, object_id = args.split(":")
        role = role.lower()
        object_id = int(object_id)

        if action == 'delete-user-role':
            user = User.query.get(object_id)
            # remove role in a subtransaction, to prevent manager shoot himself in the
            # foot
            transaction = db.session.begin_nested()
            security.ungrant_role(user, role, folder)

            if (user == g.user and role == 'manager' and
                    not has_permission(g.user, 'manage', folder, inherit=True)):

                transaction.rollback()
                flash(_('Cannot remove "manager" local role for yourself: you '
                        'don\'t have "manager" role (either by security inheritance '
                        'or by group membership)'),
                      'error')
            else:
                reindex_tree(folder)
                transaction.commit()
                flash(_("Role {role} for user {user} removed on folder {folder}"
                ).format(role=role, user=user.name, folder=folder.name),
                      "success")
        elif action == 'delete-group-role':
            group = Group.query.get(object_id)
            # remove role in a subtransaction, to prevent manager shoot himself in the
            # foot
            transaction = db.session.begin_nested()
            security.ungrant_role(group, role, folder)

            if (role == 'manager' and
                    not has_permission(g.user, 'manage', folder, inherit=True)):
                transaction.rollback()
                flash(_('Cannot remove "manager" local role for group "{group}": you'
                        ' don\'t have "manager" role by security inheritance or by '
                        'local role').format(group=group.name),
                      'error')
            else:
                flash(_("Role {role} for group {group} removed on folder {folder}"
                ).format(role=role, group=group.name, folder=folder.name),
                      "success")
                reindex_tree(folder)
                transaction.commit()

        db.session.commit()
        return redirect(
            url_for(
                ".permissions",
                folder_id=folder_id,
                community_id=folder.community.slug))


@route("/folder/<int:folder_id>/permissions_export")
@http.nocache
def permissions_export(folder_id):
    folder = repository.get_folder_by_id(folder_id)
    check_manage_access(folder)

    wb = Workbook()
    ws = wb.add_sheet("Sheet 1")

    cols = [
        ('Accès', 20),
        ('Identifiant', 40),
        ('Prénom', 14),
        ('Nom', 20),
        ('Rôle', None),
        ('Local', None),
        ('Héritage', None),
        ('Communauté', 60),
    ]

    # styling
    # from xlwt doc: width unit is 1/256 of '0' from first font in excel file
    for c, width in enumerate((c[1] for c in cols)):
        if width is not None:
            ws.col(c).width = 256 * width

    ws.row(0).height = 256 + 128

    ws.panes_frozen = True
    ws.remove_splits = True
    ws.horz_split_pos = 1

    header_style = easyxf('font: bold true;'
                          'alignment: horizontal center, vertical center;')
    for c, val in enumerate((c[0] for c in cols)):
        ws.write(0, c, val, header_style)

    # data
    permissions = iter_permissions(folder, g.user)
    row_offset = 0
    current_community = None
    for r, row in enumerate(permissions, 1):
        if current_community is None:
            current_community = row[-1]

        if current_community != row[-1]:
            current_community = row[-1]
            row_offset += 1

        # data grouping enter (refered as 'outline' in xlwt documention)
        ws.row(r + row_offset).level = 1

        for c, value in enumerate(row):
            if isinstance(value, Role):
                value = text_type(value)
            ws.write(r + row_offset, c, value)

        # data grouping exit
        ws.row(r + row_offset).level = 1

    debug = request.args.get('debug_sql')
    if debug:
        # useful only in DEBUG mode, to get the debug toolbar in browser
        return '<html><body>Exported</body></html>'

    fd = StringIO()
    wb.save(fd)

    response = make_response(fd.getvalue())
    response.headers['content-type'] = 'application/ms-excel'
    folder_name = folder.title.replace(' ', '_')
    file_date = datetime.now().strftime("%Y-%m-%d-%H:%M:%S")
    filename = 'permissions-{}-{}.xls'.format(folder_name, file_date)
    content_disposition = 'attachment;filename="{}"'.format(filename)
    response.headers['content-disposition'] = content_disposition
    return response


def iter_permissions(folder, user):
    """
    Iterator returning permissions settings on folder and its subfolders tree.
    """
    if not security.has_permission(user, "manage", folder, inherit=True):
        raise StopIteration

    community = folder.path
    local_roles = frozenset(folder.get_local_roles_assignments())
    inherited_roles = frozenset((folder.get_inherited_roles_assignments()
                                 if folder.inherit_security else []))

    result = {}
    for principal, role in (local_roles | inherited_roles):
        data = result.setdefault((principal, role), {})
        data['local'] = (principal, role) in local_roles
        data['inherit'] = (principal, role) in inherited_roles

    def _sort_key(item):
        """Sorts by name, groups first."""
        principal = item[0][0]
        is_user = isinstance(principal, User)
        item_key = [is_user]  # type: List[Any]
        if is_user:
            last_name = principal.last_name if principal.last_name else ''
            first_name = principal.first_name if principal.first_name else ''
            item_key.append(last_name.lower())
            item_key.append(first_name.lower())
        else:
            item_key.append(principal.name)
        return item_key

    for (p, role), data in sorted(result.items(), key=_sort_key):
        is_user = isinstance(p, User)
        has_access = False if is_user else '*'
        identifier = p.email if is_user else '* Group *'
        first_name = p.first_name if is_user else '-'
        last_name = p.last_name if is_user else p.name
        local = data['local']
        inherit = data['inherit']

        yield (has_access, identifier, first_name, last_name, role, local,
               inherit, community)

    subfolders = (f for f in folder.subfolders
                  if security.has_permission(user, "manage", folder))

    for subfolder in subfolders:
        for permission in iter_permissions(subfolder, user):
            yield permission


#
# Actions on folders
#
@route("/folder/<int:folder_id>", methods=['POST'])
@csrf.protect
def folder_post(folder_id):
    """
    A POST on a folder can result on several different actions (depending on the
    `action` parameter).
    """
    folder = get_folder(folder_id)
    action = request.form.get("action")

    if action == 'edit':
        return folder_edit(folder)

    elif action == 'upload':
        return upload_new(folder)

    elif action == 'download':
        return download_multiple(folder)

    elif action == 'delete':
        return delete_multiple(folder)

    elif action == 'new':
        return create_subfolder(folder)

    elif action == 'move':
        return move_multiple(folder)

    elif action == 'change-owner':
        return change_owner(folder)

    else:
        # Probably an error or a hack attempt.
        # Logger will inform sentry if enabled
        logger = logging.getLogger(__name__)
        logger.error("Unknown folder action.", extra={'stack': True})
        flash(_("Unknown action."), "error")
        return redirect(url_for(folder))


def folder_edit(folder):
    check_write_access(folder)

    changed = edit_object(folder)

    if changed:
        db.session.commit()
        flash(_("Folder properties successfully edited."), "success")
    else:
        flash(_("You didn't change any property."), "success")
    return redirect(url_for(folder))


ARCHIVE_IGNORE_FILES_GLOBS = {'__MACOSX/*', '.DS_Store'}
# translates patterns to match with any parent directory ((*/)?pattern should
# match)
ARCHIVE_IGNORE_FILES = {
    re.compile('(?:.*\\/)?' + fnmatch.translate(pattern))
    for pattern in ARCHIVE_IGNORE_FILES_GLOBS
}

# skip directory names. Directory will be created only if they contains files
ARCHIVE_IGNORE_FILES.add(re.compile(fnmatch.translate('*/')))


def explore_archive(fd, uncompress=False):
    """Given an uploaded file descriptor, return it or a list of archive
    content.

    Yield tuple(filepath, file-like object), where filepath is a list
    whose components are directories and last one is filename.
    """
    if not uncompress:
        yield [], fd
        raise StopIteration

    if is_zipfile(fd):
        with ZipFile(fd, 'r') as archive:
            for zipinfo in archive.infolist():
                filename = zipinfo.filename
                if isinstance(filename, bytes):
                    # not unicode: try to convert from utf-8 (OSX case: unicode flag not
                    # set), then legacy cp437
                    # http://stackoverflow.com/questions/13261347/correctly-decoding-zip-entry-file-names-cp437-utf-8-or
                    try:
                        filename = filename.decode('utf-8')
                    except UnicodeDecodeError:
                        filename = filename.decode('cp437')

                if any(
                        pattern.match(filename) is not None
                        for pattern in ARCHIVE_IGNORE_FILES):
                    continue

                filepath = filename.split('/')
                filename = filepath.pop()
                zip_fd = archive.open(zipinfo, 'r')
                setattr(zip_fd, 'filename', filename)
                setattr(zip_fd, 'content_type', None)
                yield filepath, zip_fd
    else:
        yield [], fd


def upload_new(folder):
    check_write_access(folder)
    session = current_app.db.session()
    base_folder = folder
    uncompress_files = 'uncompress_files' in request.form
    fds = request.files.getlist('file')
    created_count = 0
    path_cache = {}  # mapping folder path in zip: folder instance

    for upload_fd in fds:
        for filepath, fd in explore_archive(
                upload_fd, uncompress=uncompress_files):
            folder = base_folder
            parts = []
            # traverse to final directory, create intermediate if necessary. Folders
            # may be renamed if a file already exists, path_cache is used to keep
            # track of this
            for subfolder_name in filepath:
                parts.append(subfolder_name)
                path = '/'.join(parts)
                if path in path_cache:
                    folder = path_cache[path]
                    continue

                subfolders = {f.title: f for f in folder.subfolders}
                if subfolder_name in subfolders:
                    folder = subfolders[subfolder_name]
                    path_cache[path] = folder
                    continue

                subfolder_name = get_new_filename(folder, subfolder_name)
                folder = folder.create_subfolder(subfolder_name)
                session.flush()
                path_cache[path] = folder

            create_document(folder, fd)
            created_count += 1

    flash(
        _n("One new document successfully uploaded",
           "%(num)d new document successfully uploaded",
           num=created_count), "success")

    session.commit()
    return redirect(url_for(folder))


def download_multiple(folder):
    folders, docs = get_selected_objects(folder)
    if not folders:
        folders = [folder]

    def rel_path(path, content):
        return '{}/{}'.format(path, content.title)

    def zip_folder(zipfile, folder, path=''):
        for doc in folder.documents:
            doc_path = rel_path(path, doc)
            zipfile.writestr(doc_path, doc.content or b"")

        for subfolder in folder.filtered_subfolders:
            zip_folder(zipfile, subfolder, rel_path(path, subfolder))
        return zipfile

    # if using upstream send file: just create a temps file.
    # if app is streaming itself: use NamedTemporaryFile so that file is removed
    # on close()
    temp_factory = (tempfile.mktemp if current_app.use_x_sendfile else
                    tempfile.NamedTemporaryFile)
    zip_fn = temp_factory(prefix='tmp-' + current_app.name + '-', suffix='.zip')
    with ZipFile(zip_fn, "w") as zipfile:
        for doc in docs:
            zipfile.writestr(doc.title, doc.content or b"")
        for subfolder in folders:
            zip_folder(zipfile, subfolder, subfolder.title)

    if not isinstance(zip_fn, str):
        zip_fn.seek(0, os.SEEK_END)
        size = zip_fn.tell()
        zip_fn.seek(0)
    else:
        size = os.path.getsize(zip_fn)

    resp = send_file(
        zip_fn,
        mimetype="application/zip",
        as_attachment=True,
        attachment_filename=quote(folder.title.encode("utf8") + b".zip"))
    resp.headers.add('Content-Length', str(size))
    return resp


def delete_multiple(folder):
    check_write_access(folder)

    folders, docs = get_selected_objects(folder)

    for obj in docs + folders:
        app = current_app._get_current_object()
        community = g.community._model
        activity.send(
            app, actor=g.user, verb="delete", object=obj, target=community)
        repository.delete_object(obj)

    if docs + folders:
        db.session.commit()
        if docs and folders:
            msg = _("%(file_num)d files and %(folder_num)d folders sucessfully "
                    "deleted.", file_num=len(docs), folder_num=len(folders))
        elif docs and not folders:
            msg = _n(
                "1 file sucessfully deleted.",
                "%(num)d files sucessfully deleted.",
                num=len(docs))
        else:
            msg = _n(
                "1 folder sucessfully deleted.",
                "%(num)d folders sucessfully deleted.",
                num=len(folders))

        flash(msg, "success")
    else:
        flash(_("No object deleted"), "error")

    return redirect(url_for(folder))


def move_multiple(folder):
    folders, docs = get_selected_objects(folder)
    count_f = len(folders)
    count_d = len(docs)
    current_folder_url = url_for(folder)

    if not (count_f + count_d):
        flash(_('Move elements: no elements selected.'), 'info')
        return redirect(current_folder_url)

    try:
        target_folder_id = int(request.form.get('target-folder'))
    except ValueError:
        flash(_('Move elements: no destination folder selected. Aborted.'),
              'error')
        return redirect(current_folder_url)

    target_folder = repository.get_folder_by_id(target_folder_id)

    if folder == target_folder:
        flash(_('Move elements: source and destination folder are identical,'
                ' nothing done.'), 'error')
        return redirect(current_folder_url)

    if not security.has_permission(g.user, 'write', folder, inherit=True):
        # this should not happen: this is just defensive programming
        flash(_('You are not allowed to move elements from this folder'), 'error')
        return redirect(current_folder_url)

    if not security.has_permission(
            g.user, 'write', target_folder, inherit=True):
        flash(_('You are not allowed to write in folder "{folder}"'
        ).format(folder=target_folder.title),
              'error')
        return redirect(current_folder_url)

    for item in itertools.chain(folders, docs):
        # FIXME: maybe too brutal
        check_write_access(item)

    # verify we are not trying to move a folder inside itself or one of its
    # descendants
    f = target_folder
    while f:
        if f in folders:
            flash(_('Move elements: destination folder is included in moved '
                    'elements. Moved nothing.'), 'error')
            return redirect(url_for(folder))
        f = f.parent

    exist_in_dest = []
    for item in itertools.chain(folders, docs):
        try:
            with db.session.begin_nested():
                item.parent = target_folder
        except sa.exc.IntegrityError:
            exist_in_dest.append(item)

    if exist_in_dest:
        # items existing in destination: cancel operation
        db.session.rollback()
        msg = _('Move elements: canceled, some elements exists in destination '
                'folder: {elements}')
        elements = ', '.join('"{}"'.format(i.title) for i in exist_in_dest)
        flash(msg.format(elements=elements), 'error')
        return redirect(current_folder_url)

    db.session.commit()

    msg_f = (_n('1 folder', '{count} folders', count_f)
             if count_f
             else _('0 folder')).format(count=count_f)

    msg_d = (_n('1 document', '{count} documents', count_d)
             if count_d
             else _('0 document')).format(count=count_d)

    msg = _('{folders} and {documents} moved to {target}').format(
        folders=msg_f, documents=msg_d, target=target_folder.title)
    flash(msg, 'success')

    return redirect(url_for(folder))


def create_subfolder(folder):
    check_write_access(folder)

    title = request.form.get("title", "")
    description = request.form.get("description", "")
    subfolder = folder.create_subfolder(title)
    subfolder.description = description

    db.session.commit()
    return redirect(url_for(folder))


def change_owner(folder):
    check_manage_access(folder)
    items = itertools.chain(*get_selected_objects(folder))

    user_id = request.form.get('new_owner', type=int)
    user = User.query.get(user_id)

    for item in items:
        item.owner = user

    db.session.commit()
    return redirect(url_for(folder))


@route("/folder/check_valid_name")
@http.nocache
def check_valid_name():
    """Check if name is valid for content creation in this folder."""

    object_id = int(request.args.get('object_id'))
    action = request.args.get('action')
    title = request.args.get('title')

    get_object = get_document if action == 'document-edit' else get_folder
    obj = get_object(object_id)
    check_read_access(obj)

    if action == 'new':
        parent = obj
        help_text = _('An element named "{name}" is already present in folder')
    elif action in ('folder-edit', 'document-edit'):
        parent = obj.parent
        help_text = _('Cannot rename: "{name}" is already present in parent '
                      'folder')
    else:
        raise InternalServerError()

    existing = {e.title for e in parent.children}

    if action in ('folder-edit', 'document-edit'):
        try:
            existing.remove(obj.title)
        except KeyError:
            pass

    result = {}
    valid = result['valid'] = title not in existing
    if not valid:
        result['help_text'] = help_text.format(name=title)

    return jsonify(result)


@route("/folder/<int:folder_id>/descendants")
def descendants_view(folder_id):
    folder = get_folder(folder_id)
    bc = breadcrumbs_for(folder)
    actions.context['object'] = folder

    root_path_ids = folder._indexable_parent_ids + '/{}'.format(folder.id)
    index_service = get_service('indexing')

    # yapf: disable
    filters = wq.And([wq.Term('community_id', folder.community.id),
                      wq.Term('parent_ids', root_path_ids),
                      wq.Or([wq.Term('object_type', Folder.entity_type),
                             wq.Term('object_type', Document.entity_type)])])
    # yapf: enable

    results = index_service.search('', filter=filters, limit=None)
    by_path = {}
    owner_ids = set()
    for hit in results:
        by_path.setdefault(hit['parent_ids'], []).append(hit)
        owner_type, owner_id = hit['owner'].split(':')
        if owner_type == 'user':
            try:
                owner_id = int(owner_id)
                owner_ids.add(owner_id)
            except ValueError:
                pass

    for children in by_path.values():
        children.sort(
            key=lambda hit: (hit['object_type'] != Folder.entity_type, hit['name'].lower())
        )

    descendants = []

    def visit(path_id, level=0):
        children = by_path.get(path_id, ())

        for child in children:
            is_folder = child['object_type'] == Folder.entity_type
            type_letter = 'F' if is_folder else 'D'
            descendants.append((level, type_letter, child))

            if is_folder:
                path_id = child['parent_ids'] + '/{}'.format(child['id'])
                visit(path_id, level + 1)

    visit(root_path_ids, 0)

    owners = dict()
    owners_query = User.query \
        .filter(User.id.in_(owner_ids)) \
        .add_column(sa.sql.func.concat('user:', User.id).label('key'))
    for user, key in owners_query:
        owners[key] = user

    ctx = {
        'folder': folder,
        'descendants': descendants,
        'owners': owners,
        'breadcrumbs': bc,
        'get_icon': get_icon_for_hit,
    }
    return render_template("documents/descendants.html", **ctx)


def get_icon_for_hit(hit):
    if hit['object_type'] == Folder.entity_type:
        return icon_url('folder.png')

    content_type = hit['content_type']
    icon = icon_for(content_type)
    return icon



from __future__ import absolute_import, print_function, unicode_literals

from flask import url_for

from abilian.core.models.subjects import User
from abilian.sbe.testing import BaseTestCase


class TestSocialViews(BaseTestCase):

    no_login = True

    # Tests start here
    def test_home(self):
        response = self.client.get(url_for("social.home"))
        assert response.status_code == 200

    def test_users(self):
        response = self.client.get(url_for("social.users"))
        assert response.status_code == 200

    def test_groups(self):
        response = self.client.get(url_for("social.groups"))
        assert response.status_code == 200

    def test_user(self):
        user = User.query.all()[0]
        response = self.client.get(url_for("social.user", user_id=user.id))
        assert response.status_code == 200


# def test_home(client):
#     response = client.get(url_for("social.home"))
#     assert response.status_code == 200
#     # self.assert_valid(response)
#
#
# def test_users(client):
#     response = client.get(url_for("social.users"))
#     assert response.status_code == 200
#     # self.assert_valid(response)
#
#
# def test_groups(client):
#     response = client.get(url_for("social.groups"))
#     assert response.status_code == 200
#     # self.assert_valid(response)
#
#
# def test_user(client):
#     user = User.query.all()[0]
#     response = client.get(url_for("social.user", user_id=user.id))
#     assert response.status_code == 200
#     # self.assert_valid(response)



# coding=utf-8
"""
"""
from __future__ import absolute_import, print_function, unicode_literals

from datetime import datetime, timedelta

import mock
from flask_login import login_user
from pytz import UTC

from abilian.core.models.subjects import User
from abilian.sbe.apps.documents import lock
from abilian.sbe.apps.documents.lock import Lock

pytest_plugins = ['abilian.conftest']


def test_lock():
    date = datetime(2015, 10, 22, 14, 58, 42, tzinfo=UTC)
    l = Lock(user_id=3, user='Joe Smith', date=date)

    d = l.as_dict()
    assert d == dict(
        user_id=3, user='Joe Smith', date='2015-10-22T14:58:42+00:00')
    l = Lock.from_dict(d)
    assert l.user_id == 3
    assert l.user == 'Joe Smith'
    assert l.date == date


def test_lock2(app, session, test_request_context):
    user = User(
        email='test@example.com',
        first_name='Joe',
        last_name='Smith',
        can_login=True)
    other = User(email='other@exemple.com')
    session.add(user)
    session.add(other)
    session.commit()

    # set 30s lifetime
    app.config['SBE_LOCK_LIFETIME'] = 30
    dt_patcher = mock.patch.object(lock, 'utcnow', mock.Mock(wraps=lock.utcnow))
    with dt_patcher as mocked:
        created_at = datetime(2015, 10, 22, 14, 58, 42, tzinfo=UTC)
        mocked.return_value = created_at

        login_user(user)
        l = Lock.new()
        assert l.user_id == user.id
        assert l.user == 'Joe Smith'
        assert l.date == created_at
        assert l.is_owner()
        assert l.is_owner(user)
        assert not l.is_owner(other)

        assert l.lifetime == 30
        mocked.return_value = created_at + timedelta(seconds=40)
        assert l.expired

        mocked.return_value = created_at + timedelta(seconds=20)
        assert not l.expired

        login_user(other)
        assert not l.is_owner()



# coding=utf-8
"""
"""
from __future__ import absolute_import, print_function, unicode_literals

from datetime import date
from itertools import groupby, islice

import whoosh
import whoosh.query as wq
from flask import current_app, g, render_template
from flask_babel import format_date
from six import text_type
from sqlalchemy.orm import joinedload

from abilian.sbe.apps.communities.blueprint import Blueprint
from abilian.sbe.apps.documents.models import Document, icon_for
from abilian.sbe.apps.forum.models import Thread
from abilian.services import get_service
from abilian.web import url_for
from abilian.web.action import actions

from .presenters import ActivityEntryPresenter
from .util import get_recent_entries

wall = Blueprint(
    "wall", __name__, url_prefix="/wall", template_folder="templates")
route = wall.route


@wall.url_value_preprocessor
def set_current_tab(endpoint, values):
    g.current_tab = 'wall'


@route('/')
def index():
    actions.context['object'] = g.community._model
    entries = get_recent_entries(20, community=g.community)
    entries = ActivityEntryPresenter.wrap_collection(entries)
    return render_template("wall/index.html", entries=entries)


@route('/files')
def files():
    community = g.community._model
    actions.context['object'] = community

    all_files = []
    if community.has_forum:
        all_files += get_attachments_from_forum(community)

    if community.has_documents:
        all_files += get_attachments_from_dms(community)

    all_files = sorted(all_files, key=lambda doc: doc.date)
    all_files = islice(reversed(all_files), 0, 50)
    grouped_docs = group_monthly(all_files)

    return render_template("wall/files.html", grouped_docs=grouped_docs)


class Attachment(object):

    def __init__(self, url, name, owner, date, content_length, content_type):
        self.url = url
        self.name = name
        self.owner = owner
        self.date = date
        self.content_length = content_length
        self.content_type = content_type

    @property
    def icon(self):
        return icon_for(self.content_type)


def get_attachments_from_forum(community):
    all_threads = Thread.query \
        .filter(Thread.community_id == community.id) \
        .options(joinedload('posts')) \
        .options(joinedload('posts.attachments')) \
        .order_by(Thread.created_at.desc()).all()

    posts_with_attachments = []
    for thread in all_threads:
        for post in thread.posts:
            if getattr(post, 'attachments', None):
                posts_with_attachments.append(post)

    posts_with_attachments.sort(key=lambda post: post.created_at)
    posts_with_attachments.reverse()

    attachments = []
    for post in posts_with_attachments:
        for att in post.attachments:
            url = current_app.default_view.url_for(att)
            attachment = Attachment(url, att.name, text_type(att.owner),
                                    att.created_at, att.content_length,
                                    att.content_type)
            attachments.append(attachment)

    return attachments


# FIXME: significant performance issues here, needs major refactoring
def get_attachments_from_dms(community):
    index_service = get_service('indexing')
    filters = wq.And([
        wq.Term('community_id', community.id),
        wq.Term('object_type', Document.entity_type)
    ])
    sortedby = whoosh.sorting.FieldFacet('created_at', reverse=True)
    documents = index_service.search(
        '', filter=filters, sortedby=sortedby, limit=50)

    attachments = []
    for doc in documents:
        url = url_for(doc)
        attachment = Attachment(url, doc['name'], doc['owner_name'],
                                doc['created_at'], doc.get('content_length'),
                                doc.get('content_type', ''))
        attachments.append(attachment)

    return attachments


# FIXME: copy/pasted then hacked from forum/views.py
def group_monthly(objects):
    # We're using Python's groupby instead of SA's group_by here
    # because it's easier to support both SQLite and Postgres this way.
    def grouper(entity):
        return entity.date.year, entity.date.month

    def format_month(year, month):
        month = format_date(date(year, month, 1), "MMMM").capitalize()
        return "%s %s" % (month, year)

    grouped = groupby(objects, grouper)
    grouped = [(format_month(year, month), list(objs))
               for (year, month), objs in grouped]
    return grouped



# coding=utf-8
"""
Celery tasks related to document transformation and preview.
"""
from __future__ import absolute_import, division, print_function, \
    unicode_literals

import email
import mailbox
import re
from os.path import expanduser
from pathlib import Path
from typing import List, Tuple

import bleach
import chardet
from celery import shared_task
from celery.schedules import crontab
from celery.utils.log import get_task_logger
from flask import current_app, g
from flask_babel import get_locale
from flask_mail import Message
from itsdangerous import Serializer
from six import text_type

from abilian.core.celery import periodic_task
from abilian.core.extensions import db, mail
from abilian.core.models.subjects import User
from abilian.core.signals import activity
from abilian.core.util import md5
from abilian.i18n import _l, render_template_i18n
from abilian.web import url_for

from .forms import ALLOWED_ATTRIBUTES, ALLOWED_STYLES, ALLOWED_TAGS
from .models import Post, PostAttachment, Thread

MAIL_REPLY_MARKER = _l('_____Write above this line to post_____')

# logger = logging.getLogger(__package__)
# Celery logger
logger = get_task_logger(__name__)


def init_app(app):
    global check_maildir
    if app.config['INCOMING_MAIL_USE_MAILDIR']:
        make_task = periodic_task(run_every=crontab(minute='*',),)
        check_maildir = make_task(check_maildir)


@shared_task()
def send_post_by_email(post_id):
    """Send a post to community members by email.
    """
    with current_app.test_request_context('/send_post_by_email'):
        post = Post.query.get(post_id)
        if post is None:
            # deleted after task queued, but before task run
            return

        thread = post.thread
        community = thread.community
        logger.info("Sending new post by email to members of community %r",
                    community.name)

        CHUNK_SIZE = 20
        members_id = [
            member.id for member in community.members if member.can_login
        ]
        chunk = []
        for idx, member_id in enumerate(members_id):
            chunk.append(member_id)
            if idx % CHUNK_SIZE == 0:
                batch_send_post_to_users.apply_async((post.id, chunk))
                chunk = []

        if chunk:
            batch_send_post_to_users.apply_async((post.id, chunk))


@shared_task(max_retries=10, rate_limit='12/m')
def batch_send_post_to_users(post_id, members_id, failed_ids=None):
    """
    Task run from send_post_by_email; auto-retry for mails that could not be
    successfully sent.

    Task default rate limit is 6/min.: there is at least 5 seconds between 2
    batches.

    During retry, if all `members_id` fails again, the task is retried
    5min. later, then 10, 20, 40... up to 10 times before giving up. This ensures
    retries up to approximatively 3 days and 13 hours after initial attempt
    (geometric series is: 5min * (1-2**10) / 1-2) = 5115 mins).
    """
    if not members_id:
        return

    post = Post.query.get(post_id)
    if post is None:
        # deleted after task queued, but before task run
        return

    failed = set()
    successfully_sent = []
    thread = post.thread
    community = thread.community
    user_filter = (User.id.in_(members_id)
                   if len(members_id) > 1 else User.id == members_id[0])
    users = User.query.filter(user_filter).all()

    for user in users:
        try:
            with current_app.test_request_context('/send_post_by_email'):
                send_post_to_user(community, post, user)
        except:
            failed.add(user.id)
        else:
            successfully_sent.append(user.id)

    if failed:
        if failed_ids is not None:
            failed_ids = set(failed_ids)

        if failed == failed_ids:
            # 5 minutes * (2** retry count)
            countdown = 300 * 2**batch_send_post_to_users.request.retries
            batch_send_post_to_users.retry(
                [post_id, list(failed)], countdown=countdown)
        else:
            batch_send_post_to_users.apply_async([post_id, list(failed)])

    return {
        'post_id': post_id,
        'successfully_sent': successfully_sent,
        'failed': list(failed),
    }


def build_local_part(name, uid):
    """
    Build local part as 'name-uid-digest', ensuring length < 64.
    """
    tag = current_app.config['MAIL_ADDRESS_TAG_CHAR']
    key = current_app.config['SECRET_KEY']
    serializer = Serializer(key)
    signature = serializer.dumps(uid)
    digest = md5(signature)
    local_part = name + tag + uid + '-' + digest

    if len(local_part) > 64:
        if (len(local_part) - len(digest) - 1) > 64:
            # even without digest, it's too long
            raise ValueError(
                'Cannot build reply address: local part exceeds 64 '
                'characters')
        local_part = local_part[:64]

    return local_part


def build_reply_email_address(name, post, member, domain):
    # type: (text_type, Post, User, text_type) -> text_type
    """
    Build a reply-to email address embedding the locale, thread_id and user.id.

    :param name: (str)    first part of an email address
    :param post: Post()   to get post.thread_id
    :param member: User() to get user.id
    :param domain: (str)  the last domain name of the email address
    :return: (Unicode)    reply address for forum in the form
       test+P-fr-3-4-SDB7T5DXNZPD5YAHHVIKVOE2PM@testcase.app.tld

       'P' for 'post' - locale - thread id - user id - signature digest
    """
    locale = get_locale()
    uid = '-'.join(['P', str(locale), str(post.thread_id), str(member.id)])
    local_part = build_local_part(name, uid)
    return local_part + '@' + domain


def extract_email_destination(address):
    # type: (text_type) -> List[text_type]
    """Return the values encoded in the email address.

    :param address: similar to test+IjEvMy8yLzQi.xjE04-4S0IzsdicTHKTAqcqa1fE@testcase.app.tld
    :return: List() of splitted values
    """
    local_part = address.rsplit('@', 1)[0]
    tag = current_app.config['MAIL_ADDRESS_TAG_CHAR']
    name, ident = local_part.rsplit(tag, 1)
    uid, digest = ident.rsplit('-', 1)
    signed_local_part = build_local_part(name, uid)

    if local_part != signed_local_part:
        raise ValueError('Invalid signature in reply address')

    values = uid.split('-')
    header = values.pop(0)
    assert header == 'P'
    return values


def has_subtag(address):
    # type: (text_type) -> bool
    """Return True if a subtag (defined in the config.py as 'MAIL_ADDRESS_TAG_CHAR')
    was found in the name part of the address

    :param address: email adress
    """
    name = address.rsplit('@', 1)[0]
    tag = current_app.config['MAIL_ADDRESS_TAG_CHAR']
    return (tag in name)


def send_post_to_user(community, post, member):
    recipient = member.email
    subject = '[%s] %s' % (community.name, post.title)
    config = current_app.config
    sender = config.get('BULK_MAIL_SENDER', config['MAIL_SENDER'])
    SBE_FORUM_REPLY_BY_MAIL = config.get('SBE_FORUM_REPLY_BY_MAIL', False)
    SERVER_NAME = config.get('SERVER_NAME', 'example.com')
    list_id = '"{} forum" <forum.{}.{}>'.format(community.name, community.slug,
                                                SERVER_NAME)
    forum_url = url_for(
        'forum.index', community_id=community.slug, _external=True)
    forum_archive = url_for(
        'forum.archives', community_id=community.slug, _external=True)

    extra_headers = {
        'List-Id': list_id,
        'List-Archive': '<{}>'.format(forum_archive),
        'List-Post': '<{}>'.format(forum_url),
        'X-Auto-Response-Suppress': 'All',
        'Auto-Submitted': 'auto-generated',
    }

    if SBE_FORUM_REPLY_BY_MAIL and config['MAIL_ADDRESS_TAG_CHAR'] is not None:
        name = sender.rsplit('@', 1)[0]
        domain = sender.rsplit('@', 1)[1]
        replyto = build_reply_email_address(name, post, member, domain)
        msg = Message(
            subject,
            sender=sender,
            recipients=[recipient],
            reply_to=replyto,
            extra_headers=extra_headers)
    else:
        msg = Message(
            subject,
            sender=sender,
            recipients=[recipient],
            extra_headers=extra_headers)

    ctx = {
        'community': community,
        'post': post,
        'member': member,
        'MAIL_REPLY_MARKER': MAIL_REPLY_MARKER,
        'SBE_FORUM_REPLY_BY_MAIL': SBE_FORUM_REPLY_BY_MAIL,
    }
    msg.body = render_template_i18n("forum/mail/new_message.txt", **ctx)
    msg.html = render_template_i18n("forum/mail/new_message.html", **ctx)
    logger.debug("Sending new post by email to %r", member.email)

    try:
        mail.send(msg)
    except:
        logger.error(
            "Send mail to user failed",
            exc_info=True)  # log to sentry if enabled


def extract_content(payload, marker):
    """Search the payload for marker, return content up to marker."""
    index = payload.rfind(marker)
    content = payload[:index]
    return content


def validate_html(payload):
    return bleach.clean(
        payload,
        tags=ALLOWED_TAGS,
        attributes=ALLOWED_ATTRIBUTES,
        styles=ALLOWED_STYLES,
        strip=True).strip()


def add_paragraph(newpost):
    """Add surrounding <p>newpost</p> if necessary.
    """
    newpost = newpost.strip()
    if not newpost.startswith('<p>'):
        newpost = '<p>' + newpost + '</p>'
    return newpost


def clean_html(newpost):
    # type: (text_type) -> text_type
    """Clean leftover empty blockquotes."""

    clean = re.sub(
        r"(<blockquote.*?<p>.*?</p>.*?</blockquote>)",
        '',
        newpost,
        flags=re.MULTILINE | re.DOTALL)

    # this cleans auto generated reponse text (<br>timedate <a email /a><br>)
    # we reverse the string because re.sub replaces
    #  LEFTMOST NON-OVERLAPPING OCCURRENCES, and we only want the last match
    # in the string
    clean = re.sub(
        r"(>rb<.*?>a.*?=ferh\sa<.*?>rb<)",
        '',
        clean[::-1],
        flags=re.MULTILINE | re.DOTALL)
    clean = clean[::-1]

    return clean


def decode_payload(part):
    # type: (email.message.Message) -> text_type
    """Get the payload and decode (base64 & quoted printable)."""

    payload = part.get_payload(decode=True)

    if isinstance(payload, text_type):
        return payload

    charset = part.get_content_charset()
    if charset is not None:
        try:
            payload_str = payload.decode(charset)
        except UnicodeDecodeError:
            payload_str = payload.decode('raw-unicode-escape')
    else:
        # What about other encodings? -> using chardet
        found = chardet.detect(payload)
        payload_str = payload.decode(found['encoding'])

    return payload_str


def process(message, marker):
    # type: (email.message.Message, text_type) -> Tuple[text_type, List[dict]]
    """
    Check the message for marker presence and return the text up to it if present.

    :raises LookupError otherwise.
    :return: sanitized html upto marker from message and attachements
    """
    assert isinstance(message, email.message.Message)
    content = {'plain': '', 'html': ''}
    attachments = []
    # Iterate all message's parts for text/*
    for part in message.walk():
        content_type = part.get_content_type()
        content_disposition = part.get('Content-Disposition')

        if content_disposition is not None:
            attachments.append({
                'filename': part.get_filename(),
                'content_type': part.get_content_type(),
                'data': part.get_payload(decode=True),
            })

        elif content_type in ['text/plain', 'text/html']:
            subtype = part.get_content_subtype()
            payload = decode_payload(part)
            content[subtype] += payload

    if marker in content['html']:
        newpost = extract_content(content['html'], marker[:9])
        newpost = add_paragraph(validate_html(newpost))
        newpost = clean_html(newpost)
    elif marker in content['plain']:
        newpost = extract_content(content['plain'], marker[:9])
        newpost = add_paragraph(newpost)
    else:
        raise LookupError('No marker:{} in email'.format(marker))

    return newpost, attachments


@shared_task()
def process_email(message):
    # type: (email.message.Message) -> bool
    """
    Email.Message object from command line script Run message (parsed email).

    Processing chain extract community thread post member from reply_to persist
    post in db.
    """
    app = current_app._get_current_object()
    # Extract post destination from To: field, (community/forum/thread/member)
    to_address = message['To']

    if not (has_subtag(to_address)):
        logger.info('Email %r has no subtag, skipping...', to_address)
        return False

    try:
        infos = extract_email_destination(to_address)
        locale = infos[0]
        thread_id = infos[1]
        user_id = infos[2]
    except:
        logger.error(
            'Recipient %r cannot be converted to locale/thread_id/user.id',
            to_address,
            exc_info=True)
        return False

    # Translate marker with locale from email address
    rq_headers = [('Accept-Language', locale)]
    with app.test_request_context('/process_email', headers=rq_headers):
        marker = text_type(MAIL_REPLY_MARKER)

    # Extract text and attachments from message
    try:
        newpost, attachments = process(message, marker)
    except:
        logger.error('Could not Process message', exc_info=True)
        return False

    # Persist post
    with current_app.test_request_context('/process_email', headers=rq_headers):
        g.user = User.query.get(user_id)
        thread = Thread.query.get(thread_id)
        community = thread.community
        # FIXME: check membership, send back an informative email in case of an error
        post = thread.create_post(body_html=newpost)
        obj_meta = post.meta.setdefault('abilian.sbe.forum', {})
        obj_meta['origin'] = 'email'
        obj_meta['send_by_email'] = True
        activity.send(
            app, actor=g.user, verb='post', object=post, target=community)

        for desc in attachments:
            attachment = PostAttachment(name=desc['filename'])
            attachment.post = post
            attachment.set_content(desc['data'], desc['content_type'])
            db.session.add(attachment)
        db.session.commit()

    # Notify all parties involved
    send_post_by_email.delay(post.id)
    return True


def check_maildir():
    """Check the MailDir for emails to be injected in Threads.

    This task is registered only if `INCOMING_MAIL_USE_MAILDIR` is True. By
    default it is run every minute.
    """
    home = expanduser('~')
    maildirpath = str(Path(home) / 'Maildir')
    src_mdir = mailbox.Maildir(maildirpath, factory=mailbox.MaildirMessage)

    src_mdir.lock()  # Useless but recommended if old mbox is used by error

    try:
        for key, message in src_mdir.items():
            processed = process_email(message)

            # delete the message if all went fine
            if processed:
                del src_mdir[key]

    finally:
        src_mdir.close()  # Flushes all changes to disk then unlocks



# coding=utf-8

from __future__ import absolute_import, print_function, unicode_literals

from unittest import TestCase

from six import text_type

from abilian.core.entities import all_entity_classes
from abilian.core.models.subjects import Group, User

from ..models import Message, PrivateMessage


def check_editable(object):
    if hasattr(object, '__editable__'):
        for k in object.__editable__:
            assert hasattr(object, k)


def test_user():
    user = User(
        first_name="John",
        last_name="Test User",
        email="test@example.com",
        password="toto")
    check_editable(user)

    assert "John Test User" == user.name
    assert "John Test User" == text_type(user)
    # self.assertEquals(len(user.messages), 0)


def test_user_follow():
    # id is provided by DB (which is not used in this test), and is required for
    # having in (user1 != user2) == True
    user1 = User(
        id=1,
        first_name="John",
        last_name="Test User 1",
        email="test1@example.com",
        password="toto")
    user2 = User(
        id=2,
        first_name="Joe",
        last_name="Test User 2",
        email="test2@example.com",
        password="toto")

    assert len(user1.followers) == 0
    assert len(user1.followees) == 0
    assert len(user2.followers) == 0
    assert len(user2.followees) == 0

    user1.follow(user2)

    assert user1.is_following(user2)
    assert len(user2.followers) == 1
    assert len(user1.followees) == 1
    assert len(user2.followees) == 0
    assert len(user1.followers) == 0
    assert user2 in user1.followees

    user1.unfollow(user2)

    assert not user1.is_following(user2)
    assert len(user1.followers) == 0
    assert len(user1.followees) == 0
    assert len(user2.followers) == 0
    assert len(user2.followees) == 0
    assert user2 not in user1.followers


def test_group():
    user = User(
        first_name="John",
        last_name="Test User",
        email="test@example.com",
        password="toto")
    group = Group(name="Group 1")

    user.join(group)

    assert user.is_member_of(group)
    assert len(group.members) == 1
    assert len(user.groups) == 1
    assert group.members == {user}
    assert user.groups == {group}

    user.leave(group)

    assert not user.is_member_of(group)
    assert len(group.members) == 0
    assert len(user.groups) == 0


def test_private_message():
    pm = PrivateMessage(creator_id=0, recipient_id=0)
    check_editable(pm)


def test_status_update():
    user = User(
        first_name="John",
        last_name="Test User",
        email="test@example.com",
        password="toto")
    # self.assertEquals(len(user.messages), 0)

    message = Message()
    message.author = user
    check_editable(message)

    # self.assertEquals(len(user.messages), 1)
    # self.assertEquals(user.messages[0], message)
    # self.assertEquals(message.author_id, user.uid)


def test_tags():
    m = Message(content="abc #123 #cde #voilà_l_été #789")
    assert m.tags == ['123', 'cde', 'voilà_l_été', '789']


def test_get_all_entity_classes():
    classes = all_entity_classes()
    assert Message in classes
    assert PrivateMessage in classes



# coding=utf-8
"""
"""
from __future__ import absolute_import, division, print_function, \
    unicode_literals

import difflib
from pathlib import Path

import sqlalchemy as sa
from flask import current_app, flash, g, make_response, redirect, \
    render_template, request
from flask_login import current_user
from markdown import markdown
from markupsafe import Markup
from six import text_type
from six.moves.urllib.parse import quote
from sqlalchemy.orm.exc import NoResultFound
from werkzeug.exceptions import NotFound
from whoosh.searching import Hit

from abilian.core.extensions import db
from abilian.core.signals import activity
from abilian.i18n import _, _l, _n
from abilian.services.viewtracker import viewtracker
from abilian.web import csrf
from abilian.web.action import Endpoint, actions
from abilian.web.nav import BreadcrumbItem
from abilian.web.util import url_for
from abilian.web.views import ObjectCreate, ObjectEdit, ObjectView, \
    default_view

from ..communities.blueprint import Blueprint
from ..communities.common import object_viewers
from ..communities.views import default_view_kw as community_dv_kw
from .forms import WikiPageForm
from .models import WikiPage, WikiPageAttachment, WikiPageRevision

wiki = Blueprint(
    "wiki", __name__, url_prefix="/wiki", template_folder="templates")
route = wiki.route


@wiki.url_value_preprocessor
def init_wiki_values(endpoint, values):
    g.current_tab = 'wiki'

    endpoint = Endpoint('wiki.index', community_id=g.community.slug)
    g.breadcrumb.append(BreadcrumbItem(label=_l('Wiki'), url=endpoint))

    title = request.args.get('title', '').strip()
    if title and title != 'Home':
        url = Endpoint('wiki.page', community_id=g.community.slug, title=title)
        g.breadcrumb.append(BreadcrumbItem(label=title, url=url))


@route('/')
def index():
    return redirect(
        url_for(".page", title='Home', community_id=g.community.slug))


def wiki_page_default_view_kw(kw, obj, obj_type, obj_id, **kwargs):
    kw = community_dv_kw(kw, obj, obj_type, obj_id, **kwargs)
    title = kwargs.get('title')
    if title is None:
        if isinstance(obj, (Hit, dict)):
            title = obj.get('name')
        else:
            title = obj.title
        kw['title'] = title
    return kw


#
# Page level (all functions prefixed by 'page_').
#
class BasePageView(object):
    Model = WikiPage
    pk = 'title'
    Form = WikiPageForm
    base_template = 'wiki/_base.html'

    @property
    def is_home_page(self):
        return self.obj is not None and self.obj.title == 'Home'

    @property
    def template_kwargs(self):
        """
        Template render arguments. You can override `base_template` for
        instance. Only `view` and `form` cannot be overriden.
        """
        kw = super(BasePageView, self).template_kwargs
        kw['page'] = self.obj
        kw['viewers'] = object_viewers(self.obj)
        return kw

    def init_object(self, args, kwargs):
        title = kwargs['title'] = request.args['title'].strip()
        if title:
            try:
                self.obj = get_page_by_title(title)
            except NoResultFound:
                if title == 'Home':
                    self.obj = create_home_page()
                else:
                    flash(_("This page doesn't exit. You must create it first."),
                          "warning")
                    url = url_for(
                        ".page_new", title=title, community_id=g.community.slug)
                    self.redirect(url)
            actions.context['object'] = self.obj
        return args, kwargs

    def index_url(self):
        return url_for(".index", community_id=g.community.slug)

    def view_url(self):
        return url_for(
            self.view_endpoint,
            community_id=g.community.slug,
            title=self.obj.title)


class PageView(BasePageView, ObjectView):
    template = 'wiki/page.html'
    view_endpoint = '.page'
    decorators = [
        default_view(
            wiki, WikiPage, id_attr=None, kw_func=wiki_page_default_view_kw)
    ]

    def init_object(self, args, kwargs):
        args, kwargs = BasePageView.init_object(self, args, kwargs)
        if not self.obj:
            title = kwargs['title']
            if title == 'Home':
                self.obj = create_home_page()
            else:
                url = url_for(
                    ".page_edit", title=title, community_id=g.community.slug)
                return redirect(url)

        actions.context['object'] = self.obj
        viewtracker.record_hit(entity=self.obj, user=current_user)
        return args, kwargs


route('/page/')(PageView.as_view('page'))


class PageViewers(PageView):

    template = 'wiki/page_readers.html'


route('/viewers')(PageViewers.as_view('page_viewers'))


class PageEdit(BasePageView, ObjectEdit):
    # template = 'wiki/page_edit.html'
    title = _("Edit page")
    last_revision = None  # type: int
    _message_success = _l("Wiki page successfully edited.")

    def init_object(self, args, kwargs):
        if request.method != 'POST':
            return super(PageEdit, self).init_object(args, kwargs)

        page_id = request.form.get('page_id') or None
        if page_id is not None:
            page_id = int(page_id)
            self.obj = WikiPage.query.get_or_404(page_id)
            actions.context['object'] = self.obj

        return args, kwargs

    def get_form_kwargs(self):
        kwargs = ObjectEdit.get_form_kwargs(self)
        kwargs['page_id'] = kwargs['last_revision_id'] = None
        if self.obj is not None and self.obj.id:
            # if no 'id', then it's a new object (PageCreate). We shouldn't call
            # obj.last_revision, since it will issue a flush (thus creating obj and
            # the 1st revision, resulting in an "edit conflict")
            kwargs['page_id'] = self.obj.id
            kwargs['last_revision_id'] = self.obj.last_revision.id
        return kwargs

    def prepare_args(self, args, kwargs):
        super(PageEdit, self).prepare_args(args, kwargs)
        last_revision_id = self.form.last_revision_id.data
        if last_revision_id:
            self.last_revision = WikiPageRevision.query \
                .filter(WikiPageRevision.page == self.obj,
                        WikiPageRevision.id == last_revision_id) \
                .first()
        return args, kwargs

    def before_populate_obj(self):
        self.redirect_if_no_change()
        form = self.form
        self.obj.create_revision(form.body_src.data, form.message.data)

    def redirect_if_no_change(self):
        form = self.form
        if all(f.data == f.object_data for f in (form.title, form.body_src)):
            flash(_("You didn't make any change to this page."))
            return self.redirect(url_for(self.obj))

    @property
    def activity_target(self):
        return self.obj.community

    def form_invalid(self):
        is_conflict = self.form.errors.pop('last_revision_id', None)
        if not self.form.errors and is_conflict is not None:
            # no other error on form, just edit conflict: show new text
            self.form.last_revision_id.errors = []
            field = self.form.body_src
            current = self.obj.last_revision
            self.form.last_revision_id.data = current.id  # update edited revision
            self.redirect_if_no_change()  # same edition? don't bother

            if (self.last_revision.body_src == current.body_src and
                    self.form.validate()):
                # only title change? cannot show diff: save if valid
                return self.form_valid()
            else:
                edited_src = field.data
                field.data = current.body_src
                edited_diff = [
                    l for l in difflib.ndiff(
                        self.last_revision.body_src.splitlines(True),
                        edited_src.splitlines(True)) if not l[0] == '?'
                ]
                current_diff = [
                    l for l in difflib.ndiff(
                        self.last_revision.body_src.splitlines(True),
                        current.body_src.splitlines(True)) if not l[0] == '?'
                ]
                ctx = {
                    'current': current,
                    'current_diff': current_diff,
                    'edited_diff': edited_diff,
                }
                field.errors.append(
                    Markup(
                        render_template('wiki/edit_conflict_error.html',
                                        **ctx)))

        return None


route('/edit')(PageEdit.as_view('page_edit', view_endpoint='.page'))


class PageCreate(PageEdit, ObjectCreate):
    title = _l("Create page")
    _message_success = _l("Wiki page successfully created.")

    get_form_kwargs = ObjectCreate.get_form_kwargs

    def init_object(self, args, kwargs):
        args, kwargs = ObjectCreate.init_object(self, args, kwargs)
        self.obj.community = g.community
        session = sa.orm.object_session(self.obj)
        if session:
            sa.orm.session.make_transient(self.obj)
            for rev in self.obj.revisions:
                sa.orm.session.make_transient(rev)
        return args, kwargs


route('/new')(PageCreate.as_view('page_new', view_endpoint='.page'))


@route('/source/')
def page_source():
    title = request.args['title'].strip()
    try:
        page = get_page_by_title(title)
    except NoResultFound:
        return redirect(
            url_for(".page_edit", title=title, community_id=g.community.slug))

    actions.context['object'] = page
    return render_template('wiki/source.html', page=page)


@route('/changes/')
def page_changes():
    title = request.args['title'].strip()
    try:
        page = get_page_by_title(title)
    except NoResultFound:
        url = url_for(".page_edit", title=title, community_id=g.community.slug)
        return redirect(url)
    revisions = page.revisions
    revisions = sorted(revisions, key=lambda x: -x.number)
    actions.context['object'] = page
    return render_template('wiki/changes.html', page=page, revisions=revisions)


@route('/compare/')
def page_compare():
    title = request.args['title'].strip()
    try:
        page = get_page_by_title(title)
    except NoResultFound:
        return redirect(
            url_for(".page_edit", title=title, community_id=g.community.slug))
    revisions = page.revisions
    revisions = sorted(revisions, key=lambda x: x.number)
    revs_to_compare = []
    for arg in request.args:
        if arg.startswith("rev"):
            revs_to_compare.append(int(arg[3:]))
    if len(revs_to_compare) != 2:
        flash(_("You must check exactly 2 revisions."), "error")
        url = url_for(
            ".page_changes", title=title, community_id=g.community.slug)
        return redirect(url)

    revs_to_compare.sort()
    from_rev = revisions[revs_to_compare[0]]
    to_rev = revisions[revs_to_compare[1]]
    assert from_rev.number == revs_to_compare[0]
    assert to_rev.number == revs_to_compare[1]

    from_lines = from_rev.body_src.splitlines(1)
    to_lines = to_rev.body_src.splitlines(1)

    differ = difflib.Differ(charjunk=difflib.IS_CHARACTER_JUNK)
    diff = differ.compare(from_lines, to_lines)
    diff = [line for line in diff if not line.startswith("?")]

    actions.context['object'] = page
    ctx = {
        'page': page,
        'diff': diff,
        'rev1': from_rev,
        'rev2': to_rev,
    }
    return render_template('wiki/compare.html', **ctx)


@route('/delete/', methods=["POST"])
def page_delete():
    title = request.form['title'].strip()
    try:
        page = get_page_by_title(title)
    except NoResultFound:
        flash(_("This page doesn't exist"), "error")
        return redirect(url_for(".index", community_id=g.community.slug))

    db.session.delete(page)

    app = current_app._get_current_object()
    community = g.community._model
    activity.send(
        app, actor=g.user, verb="delete", object=page, target=community)

    db.session.commit()
    flash(_("Page %(title)s deleted.", title=title))
    return redirect(url_for(".index", community_id=g.community.slug))


@route('/attachments')
def attachment_download():
    title = request.args['title'].strip()
    attachment_id = int(request.args['attachment'])
    try:
        page = get_page_by_title(title)
    except NoResultFound:
        raise NotFound()

    attachment = WikiPageAttachment.query.get(attachment_id)
    assert attachment is not None
    assert attachment.wikipage is page

    response = make_response(attachment.content)
    response.headers['content-length'] = attachment.content_length
    response.headers['content-type'] = attachment.content_type
    content_disposition = ('attachment;filename="{}"'.format(
        quote(attachment.name.encode('utf8'))))
    response.headers['content-disposition'] = content_disposition
    return response


@route('/attachments', methods=['POST'])
@csrf.protect
def attachment_upload():
    title = request.args['title'].strip()
    try:
        page = get_page_by_title(title)
    except NoResultFound:
        raise NotFound()

    files = request.files.getlist('attachments')
    saved_count = 0

    for f in files:
        name = f.filename
        if not isinstance(name, text_type):
            name = text_type(f.filename, encoding='utf-8', errors='ignore')

        # FIXME: do something instead of just skipping the attachement
        if not name:
            continue

        attachment = WikiPageAttachment(name=name)
        attachment.wikipage = page
        attachment.set_content(f.read(), f.content_type)
        db.session.add(attachment)
        saved_count += 1

    if saved_count:
        db.session.commit()
        flash(
            _n("One new document successfully uploaded",
               "%(num)d new documents successfully uploaded",
               count=saved_count,
               num=len(files)), "success")
    else:
        flash(_('No file uploaded.'))

    return redirect(url_for(page))


@route('/attachments/delete', methods=['POST'])
@csrf.protect
def attachment_delete():
    title = request.args['title'].strip()
    attachment_id = int(request.args['attachment'])
    try:
        page = get_page_by_title(title)
    except NoResultFound:
        raise NotFound()

    attachment = WikiPageAttachment.query.get(attachment_id)
    assert attachment is not None
    assert attachment.wikipage is page

    if request.form.get('action') == 'delete':
        name = attachment.name
        db.session.delete(attachment)
        db.session.commit()
        flash(_('Attachment "{name}" has been deleted').format(name=name))

    return redirect(url_for(page))


#
# Wiki-level (prefixed by 'wiki_')
#
@route('/pages')
def wiki_pages():
    pages = WikiPage.query \
        .filter(WikiPage.community_id == g.community.id) \
        .order_by(WikiPage.title) \
        .all()
    return render_template('wiki/all_pages.html', pages=pages)


@route('/help')
def wiki_help():
    src = open(Path(__file__).parent / "data" / "help.txt").read()
    body = Markup(markdown(src))
    return render_template('wiki/help.html', body=body)


@route('/export')
def wiki_export():
    # TODO
    return "Not done yet"


#
# Util
#
def get_page_by_title(title):
    title = title.strip()
    page = WikiPage.query \
        .filter(WikiPage.community_id == g.community.id,
                WikiPage.title == title) \
        .one()
    return page


def create_home_page():
    path = Path(__file__).parent / 'data' / 'default_page.txt'
    default_src = path.open('rt').read()
    page = WikiPage(title="Home", body_src=default_src)
    page.community_id = g.community.id
    db.session.add(page)
    db.session.commit()
    return page



# coding=utf-8
"""
"""
from __future__ import absolute_import, print_function, unicode_literals

from datetime import datetime, timedelta

from celery import shared_task
from celery.schedules import crontab
from flask import current_app
from flask_mail import Message
from sqlalchemy import and_, or_
from validate_email import validate_email

from abilian.core.models.subjects import User
from abilian.core.util import md5
from abilian.i18n import render_template_i18n
from abilian.sbe.apps.documents.models import Document
from abilian.sbe.apps.documents.repository import repository
from abilian.sbe.apps.forum.models import Post, Thread
from abilian.sbe.apps.wiki.models import WikiPage
from abilian.services import get_service
from abilian.services.activity import ActivityEntry
from abilian.services.auth.views import get_serializer
from abilian.web import url_for

from .. import TOKEN_SERIALIZER_NAME

DIGEST_TASK_NAME = __name__ + '.send_daily_social_digest_task'
DEFAULT_DIGEST_SCHEDULE = {
    'task': DIGEST_TASK_NAME,
    'schedule': crontab(
        hour=10,
        minute=0,
    )
}


# expires after 1 day - 10 minutes
@shared_task(expires=85800)
def send_daily_social_digest_task():
    # a request_context is required when rendering templates
    with current_app.test_request_context('/send_daily_social_updates'):
        config = current_app.config
        if not config.get('PRODUCTION') or config.get("DEMO"):
            return
        send_daily_social_digest()


def send_daily_social_digest():
    for user in User.query.filter(User.can_login == True).all():
        preferences = get_service('preferences')
        prefs = preferences.get_preferences(user)

        if not prefs.get('sbe:notifications:daily', False):
            continue

        # Defensive programming.
        if not validate_email(user.email):
            continue

        try:
            send_daily_social_digest_to(user)
        except:
            current_app.logger.error(
                'Error sending daily social digest', exc_info=True)


def send_daily_social_digest_to(user):
    """Send to a given user a daily digest of activities in its communities.

    Return 1 if mail sent, 0 otherwise.
    """
    mail = current_app.extensions['mail']

    message = make_message(user)
    if message:
        mail.send(message)
        return 1
    else:
        return 0


def make_message(user):
    config = current_app.config
    sbe_config = config['ABILIAN_SBE']
    sender = config.get('BULK_MAIL_SENDER', config['MAIL_SENDER'])
    recipient = user.email
    subject = sbe_config['DAILY_SOCIAL_DIGEST_SUBJECT']
    digests = []
    happened_after = datetime.utcnow() - timedelta(days=1)
    list_id = '"{} daily digest" <daily.digest.{}>'.format(
        config['SITE_NAME'],
        config.get('SERVER_NAME', 'example.com'),
    )
    base_extra_headers = {
        'List-Id': list_id,
        'List-Post': 'NO',
        'Auto-Submitted': 'auto-generated',
        'X-Auto-Response-Suppress': 'All',
        'Precedence': 'bulk',
    }

    for membership in user.communautes_membership:
        community = membership.community
        if not community:
            # TODO: should not happen but it does. Fix root cause instead.
            continue
        # create an empty digest
        digest = CommunityDigest(community)
        AE = ActivityEntry
        activities = AE.query \
            .order_by(AE.happened_at.asc()) \
            .filter(and_(AE.happened_at > happened_after,
                 or_(and_(AE.target_type == community.object_type,
                          AE.target_id == community.id),
                     and_(AE.object_type == community.object_type,
                          AE.object_id == community.id), ))) \
            .all()

        # fill the internal digest lists with infos
        # seen_entities, new_members, new_documents, updated_documents ...
        for activity in activities:
            digest.update_from_activity(activity, user)
        # if activities:
        #   import ipdb; ipdb.set_trace()
        # save the current digest in the master digests list
        if not digest.is_empty():
            digests.append(digest)

    if not digests:
        return None

    token = generate_unsubscribe_token(user)
    unsubscribe_url = url_for(
        'notifications.unsubscribe_sbe',
        token=token,
        _external=True,
        _scheme=config['PREFERRED_URL_SCHEME'])
    extra_headers = dict(base_extra_headers)
    extra_headers['List-Unsubscribe'] = '<{}>'.format(unsubscribe_url)

    msg = Message(
        subject,
        sender=sender,
        recipients=[recipient],
        extra_headers=extra_headers)
    ctx = {
        'digests': digests,
        'token': token,
        'unsubscribe_url': unsubscribe_url
    }
    msg.body = render_template_i18n("notifications/daily-social-digest.txt",
                                    **ctx)
    msg.html = render_template_i18n("notifications/daily-social-digest.html",
                                    **ctx)
    return msg


def generate_unsubscribe_token(user):
    """Generates a unique unsubscription token for the specified user.

    :param user: The user to work with
    """
    data = [str(user.id), md5(user.password)]
    return get_serializer(TOKEN_SERIALIZER_NAME).dumps(data)


class CommunityDigest(object):

    def __init__(self, community):
        self.community = community

        self.seen_entities = set()
        self.new_members = []
        self.new_documents = []
        self.updated_documents = []
        self.new_conversations = []
        self.updated_conversations = {}
        self.new_wiki_pages = []
        self.updated_wiki_pages = {}

    def is_empty(self):
        return (not self.new_members and not self.new_documents and
                not self.updated_documents and not self.new_conversations and
                not self.updated_conversations and not self.new_wiki_pages and
                not self.updated_wiki_pages)

    def update_from_activity(self, activity, user):
        actor = activity.actor
        obj = activity.object

        # TODO ?
        # target = activity.target

        if activity.verb == 'join':
            self.new_members.append(actor)

        elif activity.verb == 'post':
            if obj is None:
                return
            if obj.id in self.seen_entities:
                return
            self.seen_entities.add(obj.id)

            if isinstance(obj, Document) and repository.has_access(user, obj):
                self.new_documents.append(obj)
            elif isinstance(obj, WikiPage):
                self.new_wiki_pages.append(obj)
            elif isinstance(obj, Thread):
                self.new_conversations.append(obj)
            elif isinstance(obj, Post):
                if obj.thread.id not in self.seen_entities:
                    # save actor and oldest/first modified Post in thread
                    # oldest post because Activities are ordered_by Asc(A.happened_at)
                    self.updated_conversations[obj.thread] = {
                        'actors': [actor],
                        'post': obj
                    }
                    # Mark this post's Thread as seen to avoid duplicates
                    self.seen_entities.add(obj.thread.id)
                elif obj.thread not in self.new_conversations:
                    # this post's Thread has already been seen in another Activity
                    # exclude it to avoid duplicates but save the Post's actor
                    self.updated_conversations[obj.thread]['actors'].append(
                        actor)

        elif activity.verb == 'update':
            if obj is None:
                return
            # special case for Wikipage, we want to know each updater
            if isinstance(obj, WikiPage):
                if obj in self.updated_wiki_pages:
                    page = self.updated_wiki_pages[obj]
                    if actor in page:
                        page[actor] += 1
                    else:
                        page[actor] = 1
                else:
                    self.updated_wiki_pages[obj] = {actor: 1}

            # fast return for all other objects
            if obj.id in self.seen_entities:
                return
            self.seen_entities.add(obj.id)

            # all objects here need to be accounted only once
            if isinstance(obj, Document) and repository.has_access(user, obj):
                self.updated_documents.append(obj)
